#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

# Note: This test is historical for the isolate_assignments attribute, which
# was deprecated and has no effect today. This test ensures it still parses
# in SystemVerilog and Verilator control files for backward compatibility.

test.scenarios('vlt')
test.top_filename = "t/t_unopt_combo.v"

test.compile(verilator_flags2=["+define+ISOLATE"])

test.execute()

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2026 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('vlt')

# --unroll-count 0 so the loops are recognized without relying on unrolling.
test.compile(verilator_flags2=['--stats', '--unroll-count', '0'])

# The leading-one positives lower to $mostsetbitp1, the count-ones positive to
# $countones; the negatives are left as loops (a wrong lowering would raise a count).
test.file_grep(test.stats,
               r'Optimizations, Loop unrolling, Lowered priority-encoder to mostsetbitp1\s+(\d+)',
               8)
test.file_grep(test.stats,
               r'Optimizations, Loop unrolling, Lowered count-set-bits to countones\s+(\d+)', 1)

test.execute()

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

# Test for bin/verilator_gantt,

import vltest_bootstrap

test.priority(30)
test.scenarios('vlt_all')
test.top_filename = "t/t_gantt.v"
test.pli_filename = "t/t_gantt_c.cpp"

test.compile(
    verilator_flags2=["--prof-exec", test.pli_filename],
    # Checks below care about thread count, so use 2 (minimum reasonable)
    threads=(2 if test.vltmt else 1))

test.execute(all_run_flags=[
    "+verilator+prof+exec+start+2",
    " +verilator+prof+exec+window+2",
    " +verilator+prof+exec+file+" + test.obj_dir + "/profile_exec.dat",
    " +verilator+prof+vlt+file+" + test.obj_dir + "/profile.vlt"])  # yapf:disable

# The profiling data goes direct to the runtime's STDOUT
#  (maybe that should go to a separate file - gantt.dat?)
gantt_log = test.obj_dir + "/gantt.log"

test.run(cmd=[
    os.environ["VERILATOR_ROOT"] + "/bin/verilator_gantt", test.obj_dir +
    "/profile_exec.dat", "--vcd " + test.obj_dir + "/profile_exec.vcd", "| tee " + gantt_log
])

if test.vltmt:
    test.file_grep(gantt_log, r'Total threads += +(\d+)', 2)
    test.file_grep(gantt_log, r'Total mtasks += +(\d+)', 6)
    # Predicted thread utilization should be less than 100%
    test.file_grep_not(gantt_log, r'Thread utilization =\s*\d\d\d+\.\d+%')
else:
    test.file_grep(gantt_log, r'Total threads += +(\d+)', 1)
    test.file_grep(gantt_log, r'Total mtasks += +(\d+)', 0)

test.file_grep(gantt_log, r'\|\s+2\s+\|\s+2\.0+\s+\|\s+eval')

# Diff to itself, just to check parsing
test.vcd_identical(test.obj_dir + "/profile_exec.vcd", test.obj_dir + "/profile_exec.vcd")

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('vlt')
test.top_filename = "t/t_protect_ids.v"

test.compile(verilator_flags2=["--protect-ids --protect-key SECRET_KEY", "t/t_protect_ids_c.cpp"])

test.execute()

# Since using a named key, we can check for always identical map
test.files_identical(test.obj_dir + "/" + test.vm_prefix + "__idmap.xml", test.golden_filename)

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('simulator')
test.top_filename = "t/t_sys_readmem.v"

# Use random reset to ensure we're fully initializing arrays before
# $writememh, to avoid miscompares with X's on 4-state simulators.
test.verilated_randReset = 2  # 2 == truly random

test.compile(v_flags2=[
    "+define+WRITEMEM_READ_BACK=1",
    "+define+WRITEMEM_BIN=1",
    '+define+WRITEMEM_READ_BACK=1',
    '\'+define+OUT_TMP1=\"' + test.obj_dir + '/tmp1.mem\"\'',
    '\'+define+OUT_TMP2=\"' + test.obj_dir + '/tmp2.mem\"\'',
    '\'+define+OUT_TMP3=\"' + test.obj_dir + '/tmp3.mem\"\'',
    '\'+define+OUT_TMP4=\"' + test.obj_dir + '/tmp4.mem\"\'',
    '\'+define+OUT_TMP5=\"' + test.obj_dir + '/tmp5.mem\"\'',
    '\'+define+OUT_TMP6=\"' + test.obj_dir + '/tmp6.mem\"\'',
    '\'+define+OUT_TMP7=\"' + test.obj_dir + '/tmp7.mem\"\'',
    '\'+define+OUT_TMP8=\"' + test.obj_dir + '/tmp8.mem\"\'',
])

test.execute()

for i in range(1, 9):
    gold = test.t_dir + "/t_sys_writemem_b.gold" + str(i) + ".mem"
    out = test.obj_dir + "/tmp" + str(i) + ".mem"
    test.files_identical(out, gold)

test.passes()



# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2026 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import os


def run(test, *, verilator_flags2=()):
    variant, fmt = test.parse_name(r"t_trace_cat_(opennext|reopen|renew)_([a-z]+)")

    # All test use the same SV file
    test.top_filename = "t/t_trace_cat.v"
    # Any variations after the format name must yield the exact same trace
    test.golden_filename = test.py_filename.rpartition(fmt)[0] + fmt + ".out"

    flags = [
        "--exe",
        f"--trace-{fmt}",
        "t/t_trace_cat.cpp",
        "-CFLAGS",
        f"-DTRACE_FMT={fmt}",
        "-CFLAGS",
        f"-DTRACE_HEADER_C=verilated_{fmt}_c.h",
        "-CFLAGS",
        f"-DVERILATED_TRACE_C=Verilated{fmt.capitalize()}C",
    ]
    match variant:
        case "opennext":
            flags.extend(["-CFLAGS", "-DTRACE_OPENNEXT"])
        case "reopen":
            flags.extend(["-CFLAGS", "-DTRACE_REOPEN"])
        case "renew":
            flags.extend(["-CFLAGS", "-DTRACE_RENEW"])
    flags.extend(verilator_flags2)

    # Run test
    test.compile(make_top_shell=False, make_main=False, verilator_flags2=flags)

    test.execute()

    if variant == "opennext":
        os.system("cat" +  \
                  " " + test.obj_dir + "/simx_part_0000.vcd" + \
                  " " + test.obj_dir + "/simx_part_0000_cat*.vcd" + \
                  " > " + test.obj_dir + "/simall.vcd")
        test.vcd_identical(test.obj_dir + "/simall.vcd", test.golden_filename)
    else:
        test.trace_identical(test.trace_filename.replace(f".{fmt}", f"_part_0000.{fmt}"),
                             test.golden_filename.replace(".out", "_part_0000.out"))
        test.trace_identical(test.trace_filename.replace(f".{fmt}", f"_part_0100.{fmt}"),
                             test.golden_filename.replace(".out", "_part_0100.out"))

    test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('vlt')
test.pli_filename = "t/t_var_pinsizes.cpp"
test.top_filename = "t/t_var_pinsizes.v"

test.compile(verilator_flags2=["-sc -pins-uint8 --trace-vcd --exe", test.pli_filename],
             make_main=False)

test.file_grep(test.obj_dir + "/" + test.vm_prefix + ".h", r'sc_core::sc_in<bool>\s+&i1;')
test.file_grep(test.obj_dir + "/" + test.vm_prefix + ".h", r'sc_core::sc_in<uint8_t>\s+&i8;')
test.file_grep(test.obj_dir + "/" + test.vm_prefix + ".h", r'sc_core::sc_in<uint16_t>\s+&i16;')
test.file_grep(test.obj_dir + "/" + test.vm_prefix + ".h", r'sc_core::sc_in<uint32_t>\s+&i32;')
test.file_grep(test.obj_dir + "/" + test.vm_prefix + ".h", r'sc_core::sc_in<uint64_t>\s+&i64;')
test.file_grep(test.obj_dir + "/" + test.vm_prefix + ".h",
               r'sc_core::sc_in<sc_dt::sc_bv<65>\s>\s+&i65;')
test.file_grep(test.obj_dir + "/" + test.vm_prefix + ".h",
               r'sc_core::sc_in<sc_dt::sc_bv<1>\s>\s+&ibv1;')
test.file_grep(test.obj_dir + "/" + test.vm_prefix + ".h",
               r'sc_core::sc_in<sc_dt::sc_bv<16>\s>\s+&ibv16;')

test.file_grep(test.obj_dir + "/" + test.vm_prefix + ".h", r'sc_core::sc_out<bool>\s+&o1;')
test.file_grep(test.obj_dir + "/" + test.vm_prefix + ".h", r'sc_core::sc_out<uint8_t>\s+&o8;')
test.file_grep(test.obj_dir + "/" + test.vm_prefix + ".h", r'sc_core::sc_out<uint16_t>\s+&o16;')
test.file_grep(test.obj_dir + "/" + test.vm_prefix + ".h", r'sc_core::sc_out<uint32_t>\s+&o32;')
test.file_grep(test.obj_dir + "/" + test.vm_prefix + ".h", r'sc_core::sc_out<uint64_t>\s+&o64;')
test.file_grep(test.obj_dir + "/" + test.vm_prefix + ".h",
               r'sc_core::sc_out<sc_dt::sc_bv<65>\s>\s+&o65;')
test.file_grep(test.obj_dir + "/" + test.vm_prefix + ".h",
               r'sc_core::sc_out<sc_dt::sc_bv<1>\s>\s+&obv1;')
test.file_grep(test.obj_dir + "/" + test.vm_prefix + ".h",
               r'sc_core::sc_out<sc_dt::sc_bv<16>\s>\s+&obv16;')

# sc_biguint pragma should override `--pins-uint8` flag
test.file_grep(test.obj_dir + "/" + test.vm_prefix + ".h",
               r'sc_core::sc_in<sc_dt::sc_biguint<1>\s>\s+&ibu1;')
test.file_grep(test.obj_dir + "/" + test.vm_prefix + ".h",
               r'sc_core::sc_in<sc_dt::sc_biguint<8>\s>\s+&ibu8;')
test.file_grep(test.obj_dir + "/" + test.vm_prefix + ".h",
               r'sc_core::sc_in<sc_dt::sc_biguint<16>\s>\s+&ibu16;')
test.file_grep(test.obj_dir + "/" + test.vm_prefix + ".h",
               r'sc_core::sc_in<sc_dt::sc_biguint<64>\s>\s+&ibu64;')
test.file_grep(test.obj_dir + "/" + test.vm_prefix + ".h",
               r'sc_core::sc_in<sc_dt::sc_biguint<512>\s>\s+&ibu512;')

test.file_grep(test.obj_dir + "/" + test.vm_prefix + ".h",
               r'sc_core::sc_out<sc_dt::sc_biguint<1>\s>\s+&obu1;')
test.file_grep(test.obj_dir + "/" + test.vm_prefix + ".h",
               r'sc_core::sc_out<sc_dt::sc_biguint<8>\s>\s+&obu8;')
test.file_grep(test.obj_dir + "/" + test.vm_prefix + ".h",
               r'sc_core::sc_out<sc_dt::sc_biguint<16>\s>\s+&obu16;')
test.file_grep(test.obj_dir + "/" + test.vm_prefix + ".h",
               r'sc_core::sc_out<sc_dt::sc_biguint<64>\s>\s+&obu64;')
test.file_grep(test.obj_dir + "/" + test.vm_prefix + ".h",
               r'sc_core::sc_out<sc_dt::sc_biguint<512>\s>\s+&obu512;')

test.execute()

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

#       irun -sv top.v t_dpi_export.v -cpost t_dpi_export_c.c -end

import vltest_bootstrap

test.scenarios('simulator')
test.top_filename = "t/t_dpi_export.v"

test.compile(
    v_flags2=["t/t_dpi_export_c.cpp"],
    verilator_flags2=["-Wall -Wno-DECLFILENAME -no-l2name", ("-O0" if test.vlt_all else "")])

test.execute()

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Test pullup/pulldown bus propagation through whole-vector ports
#
# SPDX-FileCopyrightText: 2026 Wilson Snyder
# SPDX-License-Identifier: CC0-1.0

import vltest_bootstrap

test.scenarios('simulator')

test.compile()

test.execute()

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap
import time

test.scenarios('dist')

if 'VERILATOR_TEST_RANDOM_FAILURE' not in os.environ:
    print("Test is for harness checking only, setenv VERILATOR_TEST_RANDOM_FAILURE=1")
    test.passes()
else:
    # Randomly fail to test driver.py
    t = time.time()
    if t % 2:
        test.error("random failure " + str(t))

    test.passes()



# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2026 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0


def _vcd_extract_codes(filename):
    """Parse VCD header and return dict mapping 'scope.signal' -> code."""
    codes = {}
    scope_stack = []
    with open(filename, 'r', encoding='latin-1') as fh:
        for line in fh:
            line = line.strip()
            if line.startswith("$scope"):
                parts = line.split()
                # $scope <type> <name> $end
                scope_stack.append(parts[2])
            elif line.startswith("$var"):
                parts = line.split()
                # $var <type> <width> <code> <name> ... $end
                code = parts[3]
                name = parts[4]
                full = '.'.join(scope_stack) + '.' + name
                codes[full] = code
            elif line.startswith("$upscope"):
                scope_stack.pop()
            elif line.startswith("$enddefinitions"):
                break
    return codes


def _check_aliased(test, codes, sig_a, sig_b):
    code_a = codes.get(sig_a)
    code_b = codes.get(sig_b)
    if code_a is None:
        test.error(f"Signal '{sig_a}' not found in VCD")
    if code_b is None:
        test.error(f"Signal '{sig_b}' not found in VCD")
    if code_a != code_b:
        test.error(f"Expected '{sig_a}' (code {code_a}) to alias "
                   f"'{sig_b}' (code {code_b})")


def _check_not_aliased(test, codes, sig_a, sig_b):
    code_a = codes.get(sig_a)
    code_b = codes.get(sig_b)
    if code_a is None:
        test.error(f"Signal '{sig_a}' not found in VCD")
    if code_b is None:
        test.error(f"Signal '{sig_b}' not found in VCD")
    if code_a == code_b:
        test.error(f"Expected '{sig_a}' and '{sig_b}' to have different codes, "
                   f"both have code {code_a}")


def run(test):
    (fmt, ) = test.parse_name(r"t_trace_struct_alias_([a-z]+)")

    test.top_filename = "t/t_trace_struct_alias.v"
    test.golden_filename = test.py_filename.rpartition(fmt)[0] + fmt + ".out"

    flags = ["--cc", f"--trace-{fmt}", "--trace-structs"]

    test.compile(verilator_flags2=flags)

    test.execute()

    if fmt == "vcd":
        codes = _vcd_extract_codes(test.trace_filename)

        _check_not_aliased(test, codes, "top.t.s1.a", "top.t.s2.a")

        _check_aliased(test, codes, "top.t.s3.a", "top.t.alias_of_s3a")

        _check_aliased(test, codes, "top.t.s4.a", "top.t.s5.a")
        _check_aliased(test, codes, "top.t.s4.b", "top.t.s5.b")

        _check_aliased(test, codes, "top.t.s6.a", "top.t.source_val")

    test.trace_identical(test.trace_filename, test.golden_filename)

    test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('dist')

for pl in sorted(glob.glob(test.root + "/test_regress/t/*.pl")):
    if 'bootstrap.pl' in pl:
        continue
    py = re.sub(r'\.pl', '.py', pl)
    test.error_keep_going(pl + ":1: Perl test needs conversion into a Python test '" +
                          os.path.basename(py) + "'")

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('simulator')

test.compile(
    verilator_flags2=["-unroll-count 1024", test.wno_unopthreads_for_few_cores, "--stats"])

test.execute(expect_filename=test.golden_filename)

if test.vlt:
    # Note, with vltmt this might be split differently, so only checking vlt
    test.file_grep(test.stats, r'Optimizations, Reloop iterations\s+(\d+)', 125)
    test.file_grep(test.stats, r'Optimizations, Reloops\s+(\d+)', 2)

test.passes()



# DESCRIPTION: Verilator: CI script for 'rtlmeter.yml' PR results
#
# SPDX-FileCopyrightText: 2026 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

###############################################################################
# This script runs with the venv of **RTLMeter**
###############################################################################

import sys
import json
import tabulate
from typing import Final, List

tabulate.PRESERVE_WHITESPACE = True
tabulate.MIN_PADDING = 0

_ASCII_TABLE_FORMAT: Final = tabulate.TableFormat(
    lineabove=tabulate.Line("╒═", "═", "═╤═", "═╕"),
    linebelowheader=tabulate.Line("╞═", "═", "═╪═", "═╡"),
    linebelow=tabulate.Line("╘═", "═", "═╧═", "═╛"),
    headerrow=tabulate.DataRow("│ ", " │ ", " │"),
    datarow=tabulate.DataRow("│ ", " │ ", " │"),
    linebetweenrows=None,
    padding=0,
    with_header_hide=None,
)


def printTable(table: List[List[str]], **kwargs) -> None:
    print(tabulate.tabulate(table, tablefmt=_ASCII_TABLE_FORMAT, **kwargs))


# fmt: off
stepMetric = (
    # Step,      Metric,    Status Brackets
    ("execute",  "speed",   ("❌", 0.96, "⚠️", 0.98, "✅", 1.02, "💡", 1.04, "⭐")),
    ("execute",  "memory",  ("❌", 0.90, "⚠️", 0.95, "✅", 1.05, "💡", 1.10, "⭐")),
    ("verilate", "cpu",     ("❌", 0.96, "⚠️", 0.98, "✅", 1.02, "💡", 1.04, "⭐")),
    ("verilate", "memory",  ("❌", 0.90, "⚠️", 0.95, "✅", 1.05, "💡", 1.10, "⭐")),
    ("cppbuild", "cpu",     ("❌", 0.96, "⚠️", 0.98, "✅", 1.02, "💡", 1.04, "⭐")),
    ("cppbuild", "memory",  ("❌", 0.90, "⚠️", 0.95, "✅", 1.05, "💡", 1.10, "⭐")),
    ("cppbuild", "codeSize",("❌", 0.96, "⚠️", 0.98, "✅", 1.02, "💡", 1.04, "⭐")),
    ("cppbuild", "elapsed", ("❌", 0.70, "⚠️", 0.85, "✅", 1.15, "💡", 1.30, "⭐")),
)

badgeLegend = [
    ("❌", "Likely regression"),
    ("⚠️", "Possible regression"),
    ("✅", "Within acceptable range"),
    ("💡", "Possible improvement"),
    ("⭐", "Likely improvement"),
]
# fmt: on

changedCycles = []
table = []
badgesToExplain = set()

for ref, cmp in zip(sys.argv[1::2], sys.argv[2::2]):
    with open(ref, "r", encoding="utf-8") as f:
        ref_json = json.load(f)[0]
    with open(cmp, "r", encoding="utf-8") as f:
        cmp_json = json.load(f)
    if table:
        table.append(tabulate.SEPARATING_LINE)
    runName = ref_json["runName"]

    # Check simulated cycles match - it's ok to crash if this entry does not exist
    for entry in cmp_json["execute"]["clocks"]["table"]:
        case, _, _, (refCycles, _), (newCycles, _), _, _ = entry
        refCycles = int(refCycles)
        newCycles = int(newCycles)
        if refCycles != newCycles:
            changedCycles.append([runName, case, refCycles, newCycles])

    # Check metrics
    for step, metric, brackets in stepMetric:
        if step not in cmp_json:
            continue
        data = cmp_json[step]
        if metric not in data:
            continue
        data = data[metric]
        minGain = float("inf")
        maxGain = float("-inf")
        meanGain = 1
        count = 0
        for _, _, _, (a, _), (b, _), g, _ in data["table"]:
            # for wall clock and CPU time, ignore small values that just add noise
            if metric == "elapsed" or metric == "cpu":
                if a < 30 or b < 30:
                    continue
            minGain = min(minGain, g)
            maxGain = max(maxGain, g)
            meanGain *= g
            count += 1
        if count == 0:
            continue
        meanGain = meanGain**(1 / count)

        status = brackets[0]
        for limit, badge in zip(brackets[1::2], brackets[2::2]):
            if meanGain >= limit:
                status = badge
        badgesToExplain.add(status)

        table.append([
            runName, step, ref_json["metrics"][metric]["header"], f"{meanGain:.2f}x  {status} ",
            f"{minGain:.2f}x", f"{maxGain:.2f}x", f"{count}"
        ])

# Print changed cycles if any
if changedCycles:
    print("❌ simulated cycles changed (must be fixed):")
    printTable(changedCycles,
               headers=("Run", "Case", "Old Cycles", "New Cycles"),
               colalign=("left", "left", "right", "right"),
               disable_numparse=True)
    print()

# Print results
printTable(
    table,
    headers=("Run", "Step", "Metric", "Improvement", "Min", "Max", "Samples"),
    colalign=("left", "left", "left", "right", "right", "right", "right"),
    disable_numparse=True,
)

# Explain badges
print()
for badge, legend in badgeLegend:
    if badge in badgesToExplain:
        print(f"  {badge} : {legend}")

# Fail job if status changed
sys.exit(0 if not changedCycles else 1)



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('vlt')
test.top_filename = test.obj_dir + "/t_lint_eofnewline_bad.v"


def gen(filename):
    with open(filename, 'w', encoding="utf8") as fh:  # pylint: disable=unused-variable
        pass
    # Empty file should not EOFLINE warn


gen(test.top_filename)

test.lint(verilator_flags=["-E -Wall -Wno-DECLFILENAME"], expect_filename=test.golden_filename)

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2026 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('simulator')

test.compile()

test.execute()

for filename in test.glob_some(test.obj_dir + "/" + test.vm_prefix + "___024root*.cpp"):
    test.file_grep_not(filename, r'// $c expression at')
test.file_grep_not(test.obj_dir + "/" + test.vm_prefix + "___024root__0__Slow.cpp", r'\?')
test.file_grep(test.obj_dir + "/" + test.vm_prefix + "___024root__0__Slow.cpp",
               r'VL_WRITEF_NX\(\"1.234000\\n\*-\* All Finished \*-\*\\n\"')

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('simulator')

test.compile(
    # do not test classes for multithreaded, as V3InstrCount doesn't handle MemberSel
    verilator_flags2=(['-DNO_CLASS'] if test.vltmt else []))

test.execute()

for filename in test.glob_some(test.obj_dir + "/" + test.vm_prefix + "*.cpp"):
    # Check that these simple expressions are not stored in temp variables
    test.file_grep_not(filename, r'__Vtrigcurr__expression_.* = vlSelf->clk;')
    test.file_grep_not(filename, r'__Vtrigcurr__expression_.* = vlSelf->t__DOT__q.at\(0U\);')
    test.file_grep_not(filename,
                       r'__Vtrigcurr__expression_.* = .*vlSelf->t__DOT____Vcellinp__u_array__t')
    test.file_grep_not(filename,
                       r'__Vtrigcurr__expression_.* = .*vlSymsp->TOP__t__DOT__u_class.__PVT__obj')
    # The line below should only be generated if concats/replicates aren't converted to separate senitems
    test.file_grep_not(filename, r'__Vtrigcurr__expression_.* = .*vlSelf->t__DOT__a')

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2025 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('vltmt')
test.top_filename = "t/t_hier_block_perf.v"
cycles = 100
test.sim_time = cycles * 10 + 1000

threads = 2
config_file = test.t_dir + "/" + test.name + ".vlt"
flags = [config_file, "--hierarchical", "-Wno-UNOPTFLAT", "-DSIM_CYCLES=" + str(cycles)]

test.compile(v_flags2=["--prof-pgo"] + flags, threads=threads)

test.execute(all_run_flags=[
    "+verilator+prof+exec+start+0",
    " +verilator+prof+exec+file+/dev/null",
    " +verilator+prof+vlt+file+" + test.obj_dir + "/profile.vlt"])  # yapf:disable

test.file_grep(test.obj_dir + "/profile.vlt", r'profile_data -model "VTest"')
test.file_grep(test.obj_dir + "/profile.vlt", r'profile_data -model "VCheck"')
test.file_grep(test.obj_dir + "/profile.vlt", r'profile_data -model "VCoreHier"')
test.file_grep(test.obj_dir + "/profile.vlt", r'profile_data -model "V' + test.name + '"')

# Check for cost rollovers
test.file_grep_not(test.obj_dir + "/profile.vlt", r'.*cost 64\'d\d{18}.*')

# Differentiate results
test.name = test.name + "_optimized"
test.compile(
    # Intentionally no --prof-pgo here to make sure profile data can be read in
    # without it (that is: --prof-pgo has no effect on profile_data hash names)
    v_flags2=[test.obj_dir + "/profile.vlt"] + flags,
    threads=threads)

test.execute()

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('vlt_all')
test.top_filename = "t/t_unopt_combo.v"

test.compile(v_flags2=['+define+ATTRIBUTES', "t/t_unopt_combo.vlt", "-fno-dfg"],
             # Passes, as we waived
             )

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('vlt_all', 'xsim')
test.top_filename = "t/t_lib_prot.v"

if test.benchmark:
    test.sim_time = test.benchmark * 100

trace_opt = ("" if re.search(r'--no-trace', ' '.join(test.driver_verilator_flags)) else "-trace")

# Tests the same code as t_lib_prot.py but without --protect-lib
test.compile(verilator_flags2=['--no-timing', trace_opt, "t/t_lib_prot_secret.v"],
             xsim_flags2=["t/t_lib_prot_secret.v"])

test.execute()

if test.vlt and test.trace:
    # We can see the ports of the secret module
    test.file_grep(test.trace_filename, r'accum_in')
    # and we can see what's inside (because we didn't use --protect-lib)
    test.file_grep(test.trace_filename, r'secret_')

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Primitive C++ style checker
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('dist')


def get_source_files():
    git_files = test.run_capture("cd " + test.root + " && git ls-files")
    if test.verbose:
        print("MF " + git_files)
    files = {}
    for filename in git_files.split():
        if filename == '':
            continue
        files[filename] = True
    return files


def check_pattern(filename, contents, pattern, message):
    lineno = 1
    buf = contents
    while True:
        m = re.match(r'^(.*?^(' + pattern + '))(.*)', buf)
        if not m:
            break
        lineno += m.group(1).count("\n")
        buf = m.group(3)
        test.error_keep_going(filename + ":" + str(lineno) + ": " + message)


def check_sorted(filename, contents):
    re_option = re.compile(r'^\.\. option:: (.*)')
    re_other = re.compile(r'^[^ ]')
    # .. t_dist_docs_style ignore string
    #    Ignore the given string, as a prefix match
    re_ignore = re.compile(r' *\.\. t_dist_docs_style ignore (.*)')
    # .. t_dist_docs_style restart_sort
    #    Restart the sort as if it's a new list, for forcing specific ordering
    re_restart = re.compile(r' *\.\. t_dist_docs_style restart_sort')

    contents += "__EOF__\n"

    lineno = 0
    options = []
    ignores = []
    for line in contents.split("\n"):
        lineno += 1
        if re.match(r'^($|\.\. _)', line):
            continue
        match_option = re_option.match(line)
        match_ignore = re_ignore.match(line)
        if match_option:
            arg = match_option.group(1)
            # print("-option %s" % line)
            hit = False
            for ignore in ignores:
                if arg[:len(ignore)] == ignore:
                    hit = True
            if not hit:
                options.append([lineno, arg])
        elif match_ignore:
            arg = match_ignore.group(1)
            ignores.append(arg)
        elif (options and re_other.match(line)) or re_restart.match(line):
            # print("-end-list %d %s" % (len(options), line))
            check_sorted_options(filename, options)
            ignores = []
            options = []


def check_sorted_options(filename, options):
    last_opt = None
    for opt_data in options:
        (lineno, opt) = opt_data
        if last_opt and _option_sort_key(last_opt) > _option_sort_key(opt):
            test.error_keep_going(filename + ":" + str(lineno) +
                                  ": Option '%s' should be in sorted order before '%s'" %
                                  (opt, last_opt))
        last_opt = opt


def _option_sort_key(opt):
    opt = re.sub(r'^<', ' <', opt)  # <file> before -option
    opt = re.sub(r'^\+', '-', opt)  # +options sort with -options
    opt = re.sub(r'^--', '-', opt)  # -- sorts with -
    opt = re.sub(r'^-no-', '-', opt)  # -no- sorts with non-no
    opt = opt.lower()
    return opt


#####

if not os.path.exists(test.root + "/.git"):
    test.skip("Not in a git repository")

### Must trim output before and after our file list
files = get_source_files()

for filename in sorted(files.keys()):
    filename = os.path.join(test.root, filename)
    if not os.path.exists(filename):  # git file might be deleted but not yet staged
        continue
    if not re.search(r'\.rst$', filename):
        continue

    contents = test.file_contents(filename)

    check_pattern(filename, contents, r'.*[a-z](?<!:ref):\`[^`]+\n',
                  "tag:`...` should not be split between lines")
    check_pattern(filename, contents, r'.*[a-z](?<!:ref):\'',
                  "tag:`...' should use backticks instead")
    check_sorted(filename, contents)

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.priority(50)
test.scenarios('vlt')
test.top_filename = 't/t_uvm_hello.v'

if test.have_dev_gcov:
    test.skip("Test suite intended for full dev coverage without needing this test")

test.compile(v_flags2=[
    "--binary",
    test.build_jobs_groups,
    "--CFLAGS -O0",
    "-Wall",
    "+incdir+t/uvm",  #
    "t/uvm/uvm_pkg_all_v2020_3_1_nodpi.svh",
])

test.execute(all_run_flags=['' if test.verbose else '+UVM_NO_RELNOTES'])

test.file_grep(test.run_log_filename, r'UVM TEST PASSED')

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3,
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2026 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('vlt')

# Exercise both user and generated coverage in duplicated hierarchy instances;
# the shared golden verifies inline and no-inline modules report the same
# per-hierarchy counts when forcePerInstance is used.
test.compile(top_filename="t/t_cover_hier.v",
             v_flags2=["--assert --coverage-line --coverage-user t/t_cover_hier.cpp"],
             verilator_flags2=["--exe --top-module t"],
             make_flags=['CPPFLAGS_ADD=-DTEST_OBJ_DIR="' + test.obj_dir + '"'],
             make_main=False)

test.execute()

test.files_identical_sorted(test.obj_dir + "/coverage.dat", "t/t_cover_hier.out")

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap
import shutil

test.scenarios('dist')

tmp = test.obj_dir + "/copied.dat"
shutil.copy(test.t_dir + "/t_vlcov_data_a.dat", tmp)

test.run(cmd=[
    os.environ["VERILATOR_ROOT"] + "/bin/verilator_coverage", "--unlink", tmp, "--write",
    test.obj_dir + "/output.dat"
],
         verilator_run=True)

test.files_identical(test.obj_dir + "/output.dat", "t/t_vlcov_data_a.dat")

# --unlink should have removed it
if os.path.exists(tmp):
    test.error("Wan't unlinked")

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2025 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('simulator')

test.compile(verilator_flags2=[
    "--binary", "--top cfg", "--work liba t/t_config_rules_sub.v",
    "--work libb t/t_config_rules_sub.v", "--work libc t/t_config_rules_sub.v",
    "--work libd t/t_config_rules_sub.v"
])

test.execute()

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('vlt_all')
test.top_filename = "t/t_always_reorder.v"
test.compile(verilator_flags2=["--stats -fno-reorder"])

test.file_grep(test.stats, r'Optimizations, Split always\s+(\d+)', 0)
# Here we should see some dly vars since reorder is disabled.
# (Whereas our twin test, t_alw_reorder, should see no dly vars
#  since it enables the reorder step.)
files = test.glob_some(test.obj_dir + "/" + test.vm_prefix + "___024root*.cpp")
test.file_grep_any(files, r'dly__t__DOT__v1')
test.file_grep_any(files, r'dly__t__DOT__v2')

test.execute()

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('simulator')
test.top_filename = "t/t_inst_tree.v"

out_filename = test.obj_dir + "/V" + test.name + ".tree.json"

test.compile(v_flags2=[
    "--no-json-edit-nums", "t/" + test.name + ".vlt", test.wno_unopthreads_for_few_cores
])

if test.vlt_all:
    test.file_grep(
        out_filename,
        r'{"type":"VAR","name":"u.u0.u0.z0",.*"loc":"\w,70:[^"]*",.*"origName":"z0",.*"isSigPublic":true,.*"dtypeName":"logic",.*"isSigUserRdPublic":true.*"isSigUserRWPublic":true'
    )
    test.file_grep(
        out_filename,
        r'{"type":"VAR","name":"u.u0.u0.u0.u0.z1",.*"loc":"\w,85:[^"]*",.*"origName":"z1",.*"isSigPublic":true,.*"dtypeName":"logic",.*"isSigUserRdPublic":true,.*"isSigUserRWPublic":true'
    )
    test.file_grep(
        out_filename,
        r'{"type":"VAR","name":"u.u0.u1.u0.u0.z",.*"loc":"\w,83:[^"]*",.*"origName":"z",.*,"isSigPublic":true,.*dtypeName":"logic",.*"isSigUserRdPublic":true,.*"isSigUserRWPublic":true'
    )

test.execute()
test.file_grep(test.run_log_filename, r"\] (%m|.*t\.ps): Clocked")

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2025 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('vlt')

# Compile with --binary but WITHOUT --vpi.
# Passing +verilator+vpi+ at runtime should emit a warning, not load anything.
test.compile(top_filename='t/t_flag_main.v', verilator_flags2=["--binary"])

# Without --vpi there is no loader, so the plusarg is ignored with a warning (checked via
# the golden .out).  The plusarg value is fixed, so the warning text is portable.
test.execute(all_run_flags=["+verilator+vpi+/nonexistent.so"],
             check_finished=True,
             expect_filename=test.golden_filename)

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2026 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

# +verilator+seed+0 should pick a non-zero random seed at startup, expose it
# through $get_initial_random_seed(), and pick a different value across runs
# so independent invocations are not deterministic.

test.scenarios('vlt')

test.compile(verilator_flags2=['--binary'])

seeds = []
for i in range(3):
    logfile = test.obj_dir + "/seed_zero_run" + str(i) + ".log"
    test.execute(all_run_flags=['+verilator+seed+0'], logfile=logfile)
    match = test.file_grep(logfile, r'seed=([0-9]+)')
    if match is None:
        test.error("Could not parse 'seed=NNN' from " + logfile)
    seeds.append(int(match[0]))

if any(s == 0 for s in seeds):
    test.error("+verilator+seed+0 produced a zero seed: " + repr(seeds))

if len(set(seeds)) < 2:
    test.error("+verilator+seed+0 must vary across runs, got: " + repr(seeds))

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2026 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

# Stats-variant of t_paramgraph_iface_template_nested: verifies IfaceCapture
# statistics for deeply nested interface typedef chains with clone propagation.

import vltest_bootstrap

test.scenarios('vlt')

test.top_filename = "t/t_paramgraph_iface_template_nested.v"

test.compile(v_flags2=["--binary --stats"])

test.file_grep(test.stats, r'IfaceCapture, Entries total\s+(\d+)', 25)
test.file_grep(test.stats, r'IfaceCapture, Entries template\s+(\d+)', 11)
test.file_grep(test.stats, r'IfaceCapture, Entries cloned\s+(\d+)', 14)
test.file_grep(test.stats, r'IfaceCapture, Ledger fixups in V3Param\s+(\d+)', 5)
test.file_grep(test.stats, r'IfaceCapture, Wrong-clone refs fixed\s+(\d+)', 10)
test.file_grep(test.stats, r'IfaceCapture, Dead refs fixed in modules\s+(\d+)', 0)

test.execute()

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('simulator')

child_dir = test.obj_dir + "_child"
test.mkdir_ok(child_dir)

# Compile the child
while True:
    cmdargs = test.compile_vlt_cmd(
        vm_prefix=test.vm_prefix + "_child",
        top_filename=test.name + "_child.v",
        verilator_flags=["-cc", "-Mdir", child_dir, "--debug-check"],
        # Can't use multi threading (like hier blocks), but needs to be thread safe
        threads=1)  # yapf:disable

    test.run(logfile=child_dir + "/vlt_compile.log", cmd=cmdargs)

    test.run(
        logfile=child_dir + "/vlt_gcc.log",
        cmd=[os.environ["MAKE"], "-C", child_dir,
            "-f" + os.getcwd() + "/Makefile_obj",
            "CPPFLAGS_DRIVER=-D"+test.name.upper(),
            ("CPPFLAGS_DRIVER2=-DTEST_VERBOSE=1" if test.verbose else ""),
            "VM_PREFIX=" + test.vm_prefix + "_child",
            "V" + test.name + "_child__ALL.a"  # bypass default rule, make archive
        ])  # yapf:disable

    break

# Compile the parent (might be with other than verilator)
test.compile(v_flags2=[os.path.abspath(child_dir + "/V" + test.name + "_child__ALL.a"),
                       # TODO would be nice to have this in embedded archive
                       "t/t_embed1_c.cpp"])  # yapf:disable

test.execute()

test.passes()



# DESCRIPTION: Verilator: Verilog Test bootstrap loader
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import os
import re
import sys

os.chdir(os.path.dirname(os.path.realpath(__file__)) + "/..")
# Avoid chdir leaving the .. which confuses later commands
os.environ['PWD'] = os.getcwd()
args = list(map(lambda arg: re.sub(r'.*/test_regress/', '', arg), sys.argv))
os.execl("./driver.py", "driver.py", *args)



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('vlt', 'xsim')
test.top_filename = "t/t_lib_prot.v"
if test.benchmark:
    test.sim_time = test.benchmark * 100

trace_opt = "" if re.search(r'--no-trace', ' '.join(test.driver_verilator_flags)) else "-trace"

secret_prefix = "secret"
secret_dir = test.obj_dir + "/" + secret_prefix
test.mkdir_ok(secret_dir)

while True:
    # Always compile the secret file with Verilator no matter what simulator
    #   we are testing with
    test.run(logfile=secret_dir + "/vlt_compile.log",
             cmd=["perl", os.environ["VERILATOR_ROOT"] + "/bin/verilator",
                  '--no-timing',
                  trace_opt,
                  "--prefix", "Vt_lib_prot_secret",
                  "-cc",
                  "-Mdir", secret_dir,
                  "--lib-create", secret_prefix,
                  "t/t_lib_prot_secret.v"],
             verilator_run=True)  # yapf:disable

    test.run(logfile=secret_dir + "/secret_gcc.log",
             cmd=[os.environ["MAKE"], "-C", secret_dir, "-f", "Vt_lib_prot_secret.mk"])

    test.compile(verilator_flags2=['--no-timing',
                                   trace_opt,
                                   "-LDFLAGS", secret_prefix + "/libsecret.a",
                                   secret_dir + "/secret.sv"],
                 xsim_flags2=[secret_dir + "/secret.sv"])  # yapf:disable

    test.execute(xsim_run_flags2=["--sv_lib", secret_dir + "/libsecret", "--dpi_absolute"])

    test.passes()
    break



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2025 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.priority(30)
test.scenarios('vlt_all')

test.cycles = (int(test.benchmark) if test.benchmark else 100000)
test.sim_time = test.cycles * 10 + 1000

THREADS = 2
HIER_BLOCK_THREADS = 2
HIER_THREADS = 4

config_file = test.t_dir + "/" + test.name + ".vlt"

test.compile(v_flags2=[
    config_file, "+define+SIM_CYCLES=" + str(test.cycles), "--prof-exec", "--hierarchical",
    "--stats", "-Wno-UNOPTFLAT",
    (f"-DWORKERS={HIER_BLOCK_THREADS}" if test.vltmt and HIER_BLOCK_THREADS > 1 else ""),
    (f"--hierarchical-threads {HIER_THREADS}" if test.vltmt and HIER_THREADS > 1 else "")
],
             threads=(THREADS if test.vltmt else 1),
             context_threads=(HIER_THREADS if test.vltmt else 1))

test.file_grep(test.obj_dir + "/V" + test.name + "__hier.dir/V" + test.name + "__stats.txt",
               r'Optimizations, Hierarchical DPI wrappers with costs\s+(\d+)', 6)

if test.vltmt:
    test.file_grep(test.obj_dir + "/V" + test.name + "__hier.dir/V" + test.name + "__stats.txt",
                   r'Optimizations, Thread schedule count\s+(\d+)', 2)
    test.file_grep(test.obj_dir + "/V" + test.name + "__hier.dir/V" + test.name + "__stats.txt",
                   r'Optimizations, Thread schedule total tasks\s+(\d+)', 3)

test.execute(all_run_flags=[
    "+verilator+prof+exec+start+2",
    " +verilator+prof+exec+window+2",
    " +verilator+prof+exec+file+" + test.obj_dir + "/profile_exec.dat",
    " +verilator+prof+vlt+file+" + test.obj_dir + "/profile.vlt"])  # yapf:disable

gantt_log = test.obj_dir + "/gantt.log"

test.run(cmd=[
    os.environ["VERILATOR_ROOT"] + "/bin/verilator_gantt", test.obj_dir +
    "/profile_exec.dat", "--vcd " + test.obj_dir + "/profile_exec.vcd", "| tee " + gantt_log
])

test.passes()



# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2026 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import re


def _check_empty_scopes_vcd(test, vcd_filename):
    with open(vcd_filename, "r", encoding="utf-8") as f:
        lines = f.readlines()

    scope_stack = []
    for line in lines:
        line = line.strip()
        if "$scope" in line:
            scope_stack.append({"name": line, "has_var": False})
        elif "$var" in line:
            for scope in scope_stack:
                scope["has_var"] = True
        elif "$upscope" in line:
            if scope_stack:
                scope = scope_stack.pop()
                if not scope["has_var"]:
                    test.error("Empty scope hierarchy: " + scope["name"] +
                               " (no variables in scope or any child scopes)")
        elif "$enddefinitions" in line:
            break


def _check_empty_scopes_saif(test, saif_filename):
    with open(saif_filename, "r", encoding="utf-8") as f:
        content = f.read()

    instance_stack = []
    for line in content.splitlines():
        stripped = line.strip()
        m = re.match(r"\(INSTANCE\s+(\S+)", stripped)
        if m:
            instance_stack.append({"name": m.group(1), "has_net": False})
        elif stripped.startswith("(NET"):
            for inst in instance_stack:
                inst["has_net"] = True
        elif stripped == ")" and instance_stack:
            inst = instance_stack.pop()
            if not inst["has_net"]:
                test.error("Empty INSTANCE in SAIF: " + inst["name"] +
                           " (no NET in instance or any child instances)")


def run(test):
    mode, fmt = test.parse_name(r"t_trace_struct_array_multi_inst_(cc|sc)_([a-z]+)")

    if mode == "sc" and not test.have_sc:
        test.skip("No SystemC installed")

    test.top_filename = "t/t_trace_struct_array_multi_inst.v"
    test.golden_filename = test.py_filename.rpartition(fmt)[0] + fmt + ".out"

    flags = [
        f"--{mode}",
        f"--trace-{fmt}",
        "--trace-structs",
        "--output-split-ctrace",
        "10",
    ]

    test.compile(verilator_flags2=flags)

    test.execute()

    test.trace_identical(test.trace_filename, test.golden_filename)

    if fmt in ("fst", "vcd"):
        _check_empty_scopes_vcd(test, test.golden_filename)
    elif fmt == "saif":
        _check_empty_scopes_saif(test, test.trace_filename)

    test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# SPDX-FileCopyrightText: 2026 Wilson Snyder
# SPDX-License-Identifier: CC0-1.0

import vltest_bootstrap

test.scenarios('simulator')

test.compile(timing_loop=True, verilator_flags2=['--timing'])

test.execute()

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('simulator')

test.compile(
    # Access is so we can dump waves
    v_flags2=['-trace' if test.vlt_all else ' +access+rwc'])

test.execute()

if test.vlt_all:
    test.file_grep(test.trace_filename, r'\$enddefinitions')
    sigre = re.escape("bra[ket]slash/dash-colon:9")
    test.file_grep(test.trace_filename, sigre)
    test.file_grep(test.trace_filename, r' other\.cyc ')
    test.file_grep(test.trace_filename, r' module mod\.with_dot ')
    test.vcd_identical(test.trace_filename, test.golden_filename)

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('simulator')

test.compile(verilator_flags2=['--assert --cc --coverage-user'])

test.execute()

#if test.nc: ... # See t_assert_cover.py for NC version

# Allow old SystemC::Coverage format dump, or new binary dump
# Check that the hierarchy doesn't include __PVT__
# Otherwise our coverage reports would look really ugly
if test.vlt_all:
    test.file_grep(test.coverage_filename, r'(top\.t\.sub.*.cyc_eq_5)')

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('linter')

test.compile(verilator_flags2=['--assert'],
             nc_flags2=['+assert'],
             vcs_flags2=['-assert svaext'],
             fails=True,
             expect_filename=test.golden_filename)

test.extract(in_filename=test.top_filename,
             out_filename=test.root + "/docs/gen/ex_USERWARN_faulty.rst",
             regexp=r'\$warn.*User')

test.extract(in_filename=test.top_filename,
             out_filename=test.root + "/docs/gen/ex_USERERROR_faulty.rst",
             regexp=r'\$error.*User')

test.extract(in_filename=test.top_filename,
             out_filename=test.root + "/docs/gen/ex_USERINFO_faulty.rst",
             regexp=r'\$info.*User')

test.extract(in_filename=test.top_filename,
             out_filename=test.root + "/docs/gen/ex_USERFATAL_faulty.rst",
             regexp=r'\$fatal.*User')

test.extract(in_filename=test.golden_filename,
             out_filename=test.root + "/docs/gen/ex_USERWARN_msg.rst",
             regexp=r'USERWARN:.* User')

test.extract(in_filename=test.golden_filename,
             out_filename=test.root + "/docs/gen/ex_USERERROR_msg.rst",
             regexp=r'USERERROR:.* User')

test.extract(in_filename=test.golden_filename,
             out_filename=test.root + "/docs/gen/ex_USERINFO_msg.rst",
             regexp=r'-Info:.* User')

test.extract(in_filename=test.golden_filename,
             out_filename=test.root + "/docs/gen/ex_USERFATAL_msg.rst",
             regexp=r'USERFATAL:.* User')

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

# Test for bin/verilator_gantt,

import vltest_bootstrap

test.priority(50)
test.scenarios('vlt_all')
test.top_filename = "t/t_gantt.v"
test.pli_filename = "t/t_gantt_c.cpp"

threads_num = (2 if test.vltmt else 1)

test.compile(
    make_top_shell=False,
    make_main=False,
    verilator_flags2=["--prof-exec --exe", test.pli_filename, "t/t_gantt_two.cpp"],
    # Checks below care about thread count, so use 2 (minimum reasonable)
    threads=threads_num,
    make_flags=["CPPFLAGS_ADD=\"-DVL_NO_LEGACY -DTEST_USE_THREADS=" + str(threads_num) + "\""])

test.execute(all_run_flags=[
    "+verilator+prof+exec+start+4",
    " +verilator+prof+exec+window+4",
    " +verilator+prof+exec+file+" + test.obj_dir + "/profile_exec.dat",
    " +verilator+prof+vlt+file+" + test.obj_dir + "/profile.vlt"])   # yapf:disable

gantt_log = test.obj_dir + "/gantt.log"

# The profiling data goes direct to the runtime's STDOUT
#  (maybe that should go to a separate file - gantt.dat?)
test.run(cmd=[
    os.environ["VERILATOR_ROOT"] + "/bin/verilator_gantt",
    test.obj_dir + "/profile_exec.dat",
    "--vcd " + test.obj_dir + "/profile_exec.vcd",
    "| tee " + gantt_log])   # yapf:disable

if test.vltmt:
    test.file_grep(gantt_log, r'Total threads += +(\d+)', 2)
    test.file_grep(gantt_log, r'Total mtasks += +(\d+)', 6)
else:
    test.file_grep(gantt_log, r'Total threads += +(\d+)', 1)
    test.file_grep(gantt_log, r'Total mtasks += +(\d+)', 0)

test.file_grep(gantt_log, r'\|\s+4\s+\|\s+4\.0+\s+\|\s+eval')

# Diff to itself, just to check parsing
test.vcd_identical(test.obj_dir + "/profile_exec.vcd", test.obj_dir + "/profile_exec.vcd")

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2025 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('vltmt')
test.clean_objs()
test.top_filename = "t/t_dpi_instr_count_large.v"

test.compile(
    v_flags2=["t/t_dpi_instr_count_large.cpp"],
    verilator_flags2=[
        "--hierarchical",
        "--instr-count-dpi 999999999",
        # Force UNOPTTHREADS error to cause Contraction limit increase beyond UINT32
        "--threads-max-mtasks 1",
        "-Wno-UNOPTTHREADS"
    ],
    threads=2)

test.execute()

test.passes()



# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2026 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import os


def run(test, *, verilator_flags2=()):
    variant, mode, fmt = test.parse_name(r"t_trace_two_(dump|port|hdr)_(cc|sc)_([a-z]+).*")

    if mode == "sc" and not test.have_sc:
        test.skip("No SystemC installed")

    # All test use the same SV file
    test.top_filename = "t/t_trace_two_a.v"
    # Main driver file depends on mode
    main_filename = f"t/t_trace_two_{mode}.cpp"
    # Any variations after the format name must yield the exact same trace
    test.golden_filename = test.py_filename.rpartition(fmt)[0] + fmt + ".out"

    common_flags = [
        f"--{mode}",
        f"--trace-{fmt}",
        f"+define+TRACE_FMT={fmt}",
        "-CFLAGS",
        f"-DTRACE_FMT={fmt}",
    ]
    match mode:
        case "cc":
            common_flags.extend([
                "-CFLAGS",
                f"-DTRACE_HEADER_C=verilated_{fmt}_c.h",
                "-CFLAGS",
                f"-DVERILATED_TRACE_C=Verilated{fmt.capitalize()}C",
            ])
        case "sc":
            common_flags.extend([
                "-CFLAGS",
                f"-DTRACE_HEADER_SC=verilated_{fmt}_sc.h",
                "-CFLAGS",
                f"-DVERILATED_TRACE_SC=Verilated{fmt.capitalize()}Sc",
            ])
        case _:
            test.error(f"Unhandled test mode '{mode}'")
    match variant:
        case "dump":
            common_flags.append("+define+TEST_DUMP")
        case "port":
            common_flags.append("+define+TEST_DUMPPORTS")
        case "hdr":
            common_flags.extend(["-CFLAGS", "-DTEST_HDR_TRACE=1"])
        case _:
            test.error(f"Unhandled test variant '{variant}'")
    common_flags.extend(verilator_flags2)

    # Verilate 'b'
    test.compile(make_main=False,
                 verilator_make_gmake=False,
                 top_filename='t_trace_two_b.v',
                 vm_prefix='Vt_trace_two_b',
                 verilator_flags2=common_flags)

    # Create the ALL.cpp for 'b'
    test.run(logfile=test.obj_dir + "/make_first_ALL.log",
             cmd=[
                 os.environ["MAKE"], "-C", "" + test.obj_dir, "-f", "Vt_trace_two_b.mk",
                 "Vt_trace_two_b__ALL.cpp"
             ])

    # Build 'a', this includes ALL.cpp for 'b'
    test.compile(make_main=False,
                 top_filename='t_trace_two_a.v',
                 verilator_flags2=common_flags + ['--exe', main_filename])

    test.execute()

    test.trace_identical(test.trace_filename, test.golden_filename)

    test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('simulator')
test.top_filename = "t/t_flag_main_top_name.v"

test.compile(verilator_flags=["-Mdir " + test.obj_dir, "--exe", "--build", "--main"],
             verilator_flags2=["--top-module top", "--main-top-name -", "-DMAIN_TOP_NAME_EMPTY"],
             make_main=False)

test.execute()

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2026 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('simulator')

test.top_filename = "t_to_string_emitted.v"

test.compile(verilator_flags2=["--stats"])

test.execute()

test.file_grep(test.stats, r"Optimizations, Struct/union ToString emitted\s+(\d+)", 0)
test.file_grep(test.stats, r"Optimizations, Class ToString emitted\s+(\d+)", 0)
test.file_grep(test.stats, r"Optimizations, Interface ToString emitted\s+(\d+)", 0)

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap
import glob

test.scenarios("vlt")

test.compile(verilator_flags2=["--x-initial unique"])

test.execute(all_run_flags=["+verilator+rand+reset+2"], expect_filename=test.golden_filename)

test.passes()

other_logs = [x for x in glob.glob("t/t_x_rand_stability_*.out") if "_zero" not in x]
for other_log in other_logs:
    test.files_identical(test.golden_filename, other_log)



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('vlt')
test.top_filename = "t/t_file_does_not_exist.v"

# Tests for the error message and then the absence of the
# "Command Failed" line
test.compile(v_flags2=["--quiet-exit"], fails=True)

test.file_grep_not(test.compile_log_filename, r'Exiting due to')
test.file_grep_not(test.compile_log_filename, r'Command Failed')

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('simulator')
test.top_filename = "t/t_inst_tree.v"

out_filename = test.obj_dir + "/V" + test.name + ".tree.json"

test.compile(v_flags2=["--no-json-edit-nums", test.t_dir + "/" + test.name + ".vlt"])

if test.vlt_all:
    test.file_grep(out_filename,
                   r'{"type":"MODULE","name":"l1",.*"loc":"\w,56:[^"]*",.*"origName":"l1"')
    test.file_grep(out_filename,
                   r'{"type":"MODULE","name":"l2",.*"loc":"\w,62:[^"]*",.*"origName":"l2"')
    test.file_grep(out_filename,
                   r'{"type":"MODULE","name":"l3",.*"loc":"\w,69:[^"]*",.*"origName":"l3"')
    test.file_grep(out_filename,
                   r'{"type":"MODULE","name":"l4",.*"loc":"\w,76:[^"]*",.*"origName":"l4"')
    test.file_grep(out_filename,
                   r'{"type":"MODULE","name":"l5__P1",.*"loc":"\w,83:[^"]*",.*"origName":"l5"')
    test.file_grep(out_filename,
                   r'{"type":"MODULE","name":"l5__P2",.*"loc":"\w,83:[^"]*",.*"origName":"l5"')

test.execute()
test.file_grep(test.run_log_filename, r"\] (%m|.*t\.ps): Clocked")

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2026 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('linter')

# Issue #5066: Nested interface ports through interface arrays
# (e.g., l2.l1[0] where l1 is an interface array inside interface l2).
# V3Param internal errors have been fixed, but V3LinkDot interface
# connection resolution for array element selections is not yet implemented.
# This test documents the current behavior and should be updated when
# full support is added.
test.lint(fails=True, expect_filename=test.golden_filename)

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# Copyright 2025 by Wilson Snyder. This program is free software; you
# can redistribute it and/or modify it under the terms of either the GNU
# Lesser General Public License Version 3 or the Perl Artistic License
# Version 2.0.
# SPDX-FileCopyrightText: 2025 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('vlt')

test.compile(verilator_flags2=["--binary"])

test.execute()

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: generic coverage declaration dump test
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2026 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

from pathlib import Path

import vltest_bootstrap

test.scenarios('vlt')
test.top_filename = "t/t_cover_fsm_policy_accept_multi.v"

# Dump generic COVEROTHERDECL nodes so AstCoverOtherDecl::dump() also sees
# coverage declarations with no FSM metadata, exercising the empty-field side
# of the fv/ff/ft/fg formatting.
test.lint(v_flags=["--coverage-line", "--dump-tree"])

tree_files = [Path(filename) for filename in test.glob_some(test.obj_dir + "/*.tree")]
tree_texts = [filename.read_text(encoding="utf8") for filename in tree_files]

generic_lines = []
for text in tree_texts:
    generic_lines.extend(line for line in text.splitlines()
                         if "COVEROTHERDECL" in line and " page=v_line/" in line)

assert generic_lines
assert any(" fv=" not in line and " ff=" not in line and " ft=" not in line and " fg=" not in line
           for line in generic_lines)

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('vltmt')
test.top_filename = "t/t_gen_alw.v"  # It doesn't really matter what test

test.compile(v_flags2=["--prof-pgo"], threads=2)

test.execute(all_run_flags=[
    "+verilator+prof+exec+start+0",
    " +verilator+prof+exec+file+/dev/null",
    " +verilator+prof+vlt+file+" + test.obj_dir + "/profile.vlt"])  # yapf:disable

test.file_grep(test.obj_dir + "/profile.vlt", r'profile_data ')

test.compile(
    # Intentionally no --prof-pgo here to make sure profile data can be read in
    # without it (that is: --prof-pgo has no effect on profile_data hash names)
    v_flags2=[" " + test.obj_dir + "/profile.vlt"],
    threads=2)

test.execute()

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3,
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2026 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

from coverage_common import init_log, run_vlcov, vlcov_run_context

test.scenarios('dist')

log = test.obj_dir + "/vlcov.log"
tmp_log = test.obj_dir + "/vlcov.tmp"

init_log(log)
vlcov = vlcov_run_context(test, log, tmp_log)

collapsed_cov = test.obj_dir + "/collapsed.dat"
with open(collapsed_cov, "w", encoding="utf-8") as fh:
    fh.write("# SystemC::Coverage-3\n")
    fh.write("C '\001f\002t/soc.v\001l\00210\001t\002line\001page\002v_line/core"
             "\001h\002tb.cluster*.u_core' 4\n")
    fh.write("C '\001f\002t/soc.v\001l\00211\001t\002line\001page\002v_line/core"
             "\001h\002tb.cluster*.u_core' 0\n")
    fh.write("C '\001f\002t/soc.v\001l\00212\001t\002branch\001page\002v_branch/core"
             "\001h\002tb.cluster*.u_core' 9\n")
    fh.write("C '\001f\002t/soc.v\001l\00213\001t\002branch\001page\002v_branch/core"
             "\001h\002tb.cluster?.u_core' 0\n")

empty_cov = test.obj_dir + "/empty.dat"
with open(empty_cov, "w", encoding="utf-8") as fh:
    fh.write("# SystemC::Coverage-3\n")

run_vlcov(vlcov,
          "verilator_coverage --report summary t/t_vlcov_data_e.dat",
          args=["--report", "summary", "t/t_vlcov_data_e.dat"])
run_vlcov(vlcov, "verilator_coverage t/t_cover_hier.out", args=["t/t_cover_hier.out"])
run_vlcov(vlcov,
          "verilator_coverage --report summary,hier t/t_cover_hier.out",
          args=["--report", "summary,hier", "t/t_cover_hier.out"])
run_vlcov(vlcov,
          "verilator_coverage --report hier,summary --levels 0 t/t_cover_hier.out",
          args=["--report", "hier,summary", "--levels", "0", "t/t_cover_hier.out"])
run_vlcov(vlcov,
          "verilator_coverage --report hier,summary --levels -4 t/t_cover_hier.out",
          args=["--report", "hier,summary", "--levels", "-4", "t/t_cover_hier.out"])
run_vlcov(vlcov,
          "verilator_coverage --report summary empty.dat",
          args=["--report", "summary", empty_cov])
run_vlcov(vlcov,
          "verilator_coverage --report hierarchy t/t_cover_hier.out",
          args=["--report", "hierarchy", "t/t_cover_hier.out"])
run_vlcov(vlcov,
          "verilator_coverage --report hierarchy --levels 2 t/t_cover_hier.out",
          args=["--report", "hierarchy", "--levels", "2", "t/t_cover_hier.out"])
run_vlcov(vlcov,
          "verilator_coverage --report hierarchy --levels -1 t/t_cover_hier.out",
          args=["--report", "hierarchy", "--levels", "-1", "t/t_cover_hier.out"])
run_vlcov(vlcov,
          "verilator_coverage --report hier --levels 2 t/t_cover_hier.out",
          args=["--report", "hier", "--levels", "2", "t/t_cover_hier.out"])
run_vlcov(vlcov,
          "verilator_coverage --report hierarchy collapsed.dat",
          args=["--report", "hierarchy", collapsed_cov])

test.files_identical(log, test.golden_filename)

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap
import platform

test.scenarios('vlt_all')

if re.search(r'clang', test.cxx_version) and 'aarch64' in platform.processor():
    test.skip("Known compiler profile issues on clang aarch64")
if platform.libc_ver()[0] != "glibc":
    test.skip("The test depends on GMON_OUT_PREFIX which is glibc-specific")

test.compile(verilator_flags2=["--stats --prof-cfuncs +define+T_PROF"])

# TODO below might no longer be required as configure checks for -pg
if 'VERILATOR_TEST_NO_GPROF' in os.environ:
    test.skip("Skipping due to VERILATOR_TEST_NO_GPROF")

for filename in glob.glob(test.obj_dir + "/gmon.out.*"):
    test.unlink_ok(filename)
test.setenv('GMON_OUT_PREFIX', test.obj_dir + "/gmon.out")

test.execute()

gmon_path = None
for filename in glob.glob(test.obj_dir + "/gmon.out.*"):
    gmon_path = filename
if not gmon_path:
    test.error("Profiler did not create a gmon.out")
gmon_base = re.sub(r'.*[/\\]', '', gmon_path)

test.run(
    cmd=["cd " + test.obj_dir + " && gprof " + test.vm_prefix + " " + gmon_base + " > gprof.log"],
    check_finished=False)

test.run(cmd=[
    "cd " + test.obj_dir + " && " + os.environ["VERILATOR_ROOT"] +
    "/bin/verilator_profcfunc gprof.log > profcfuncs.log"
],
         check_finished=False)

test.file_grep(test.obj_dir + "/profcfuncs.log", r'Overall summary by')
test.file_grep(test.obj_dir + "/profcfuncs.log", r'VLib + VL_POWSS_QQQ')
test.file_grep(test.obj_dir + "/profcfuncs.log", r'VBlock + t_prof:')

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: FSM coverage for non-inlined fsm_register_wrapper state register wrappers
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2026 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import os

import vltest_bootstrap

test.scenarios('simulator')

test.compile(
    verilator_flags2=['--cc --coverage-fsm -fno-inline t/t_fsm_register_wrapper_noinline.vlt'])

test.execute()

test.run(
    cmd=[
        os.environ["VERILATOR_ROOT"] + "/bin/verilator_coverage",
        "--annotate",
        test.obj_dir + "/annotated",
        test.obj_dir + "/coverage.dat",
    ],
    verilator_run=True,
)

test.files_identical(test.obj_dir + "/annotated/" + test.name + ".v", test.golden_filename)

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2026 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('simulator')

test.top_filename = "t/t_procedure_nba.v"

test.compile(verilator_flags2=["--binary", "--stats", "-DPROCEDURE=initial"])

test.file_grep(test.stats, r'Scheduling, \'act\' extra triggers\s+(\d+)', 0)
test.file_grep(test.stats, r'Scheduling, \'act\' pre triggers\s+(\d+)', 0)
test.file_grep(test.stats, r'Scheduling, \'act\' sense triggers\s+(\d+)', 3)
test.file_grep(test.stats, r'Procedures needing initial NBA trigger\s+(\d+)', 100)

test.execute()

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Runtime hierarchy report with per-instance coverage data
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3,
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2026 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

from coverage_common import init_log, run_vlcov, vlcov_run_context

test.scenarios('simulator')

if not test.have_coroutines:
    test.skip("Test requires Coroutines; ignore error since not available")

test.compile(top_filename="t/t_vlcov_hier_report_runtime.v",
             verilator_flags2=[
                 "--binary",
                 "--coverage",
                 "--coverage-fsm",
                 "--coverage-per-instance",
                 "--top-module",
                 "tb",
                 "--timing",
             ])

test.execute(all_run_flags=[" +verilator+coverage+file+" + test.obj_dir + "/coverage.dat"])

summary_log = test.obj_dir + "/summary.log"
hier_log = test.obj_dir + "/hierarchy.log"
combined_log = test.obj_dir + "/vlcov.log"

init_log(combined_log)
run_vlcov(vlcov_run_context(test, combined_log, summary_log),
          "verilator_coverage --report summary coverage.dat",
          args=["--report", "summary", test.obj_dir + "/coverage.dat"])
run_vlcov(vlcov_run_context(test, combined_log, hier_log),
          "verilator_coverage --report hierarchy --levels 3 coverage.dat",
          args=["--report", "hierarchy", "--levels", "3", test.obj_dir + "/coverage.dat"])

test.files_identical(combined_log, test.golden_filename)

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('simulator_st')
test.top_filename = "t/t_assert_elab.v"

test.unlink_ok(test.obj_dir + "/t_assert_elab_bad.log")

test.compile(v_flags2=[
    '+define+FAILING_ASSERTIONS', ('--assert' if test.vlt_all else ('+assert' if test.nc else ''))
],
             fails=True)

test.file_grep(test.compile_log_filename,
               r'%Warning-USERFATAL: "Parameter   5 is invalid...string and constant both work"')

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import json
import vltest_bootstrap

test.scenarios('vlt')
test.top_filename = "t/t_dump.v"

out_filename = test.obj_dir + "/V" + test.name + "_002_cellsort.tree.json"

test.lint(v_flags=["--dump-tree-json --no-json-edit-nums"])

# not files_identical as output is too sensitive to internal changes
test.file_grep(out_filename, r'"type":"MODULE"')

# check JSON is parsable
with open(out_filename, 'r', encoding="utf8") as fh:
    json.load(fh)

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2025 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('simulator')

test.compile(verilator_flags2=[
    '--binary', '--top cfg1', '--work liba', 't/t_config_work__liba.v', '--work libb',
    't/t_config_work__libb.v'
])

test.execute()

# Sort so that 'initial' scheduling order is not relevant
test.files_identical_sorted(test.run_log_filename, test.golden_filename, is_logfile=True)

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('simulator')
test.top_filename = test.obj_dir + "/t_vlprocess_missing.v"

# Number of tests to generate
NUM_TESTS = 200

# Testbench header template
HEADER = """\
module Testbench;

  logic clk;
  logic reset;

  // Clock driver
  initial begin
    clk = 0;
    forever begin
      #5 clk = ~clk;
    end
  end

  task automatic advance_clock(int n = 1);
    repeat (n) @(posedge clk);
  endtask

"""

# Test task template
TEST_TASK_TEMPLATE = """
  task automatic test_{num}();
    int counter = 0;
    int expected_value = {num};

    // Timeout wait
    fork
      begin
        advance_clock(10000);
        $error("Timeout");
      end
    join_none
    wait (counter == expected_value);
    disable fork;

    while (counter < expected_value) begin
      advance_clock();
      counter++;
    end
  endtask
"""

# Testbench footer template
FOOTER = "  initial begin"

# Call template for invoking each test task
CALL_TEMPLATE = "    test_{num}();\n"

# Footer end
FOOTER_END = """
    $finish;
  end

endmodule
"""


def gen(filename, num_tests):
    """
    Generates a SystemVerilog testbench with the specified number of tests.

    Args:
        filename (str): The output file name for the generated testbench.
        num_tests (int): The number of test tasks to generate.
    """
    with open(filename, 'w', encoding="utf-8") as fh:
        fh.write("// Generated by t_vlprocess_missing.py\n")

        # Write the header
        fh.write(HEADER)

        # Generate the test tasks
        for i in range(1, num_tests + 1):
            fh.write(TEST_TASK_TEMPLATE.format(num=i))

        # Write the initial block with test calls
        fh.write(FOOTER)
        for i in range(1, num_tests + 1):
            fh.write(CALL_TEMPLATE.format(num=i))
        fh.write(FOOTER_END)


gen(test.top_filename, NUM_TESTS)

test.compile(verilator_flags2=["--binary"])

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2026 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('dist')

# Real-world repro: a misconfigured VERILATOR_BIN that points back at the
# Perl wrapper used to fork-bomb the host. The wrapper now caps re-entry
# depth via $VERILATOR_RUNNING and aborts past the cap.
os.environ['VERILATOR_BIN'] = os.environ["VERILATOR_ROOT"] + "/bin/verilator"

# --no-unlimited-stack avoids the ulimit_stack_unlimited() backtick branch,
# which would otherwise double the recursion fanout. Linear re-entry through
# run() is enough to exercise the depth-cap abort.
test.run(
    fails=True,
    cmd=[os.environ["VERILATOR_ROOT"] + "/bin/verilator", "--no-unlimited-stack", "--version"],
    logfile=test.run_log_filename)

test.file_grep(test.run_log_filename,
               r'%Error: verilator: re-entered \d+ levels deep via \$VERILATOR_RUNNING')

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: FSM coverage negative extraction test
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2026 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import os

import vltest_bootstrap

test.scenarios('simulator')

# This test is intentionally "half supported": one case item is a simple
# direct state assignment, while the others use shapes the extractor should
# ignore (multiple meaningful statements or assignment to a non-state lhs).
# That lets us hit the conservative negative branches in directStateAssign()
# and singleMeaningfulStmt() without changing user-visible behavior.
test.compile(verilator_flags2=['--cc --coverage-fsm'])

test.execute()

# Use annotated-source output so the golden locks down which candidate arcs
# survive extraction and which unsupported shapes are intentionally skipped.
test.run(cmd=[
    os.environ["VERILATOR_ROOT"] + "/bin/verilator_coverage",
    "--annotate",
    test.obj_dir + "/annotated",
    test.obj_dir + "/coverage.dat",
],
         verilator_run=True)

annotated_filename = test.obj_dir + "/annotated/" + test.name + ".v"
normalized_filename = test.obj_dir + "/annotated/" + test.name + ".normalized.v"
with open(annotated_filename, encoding="utf-8") as in_file:
    with open(normalized_filename, "w", encoding="utf-8") as out_file:
        for line in in_file:
            out_file.write(line.rstrip() + "\n")

test.files_identical(normalized_filename, test.golden_filename)

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Hacky import order checker, used to ensure all getters
# come before setters for consistent codegen when using autocxx (#5182)
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('dist')


def get_source_files():
    git_files = test.run_capture("cd " + test.root + " && git ls-files")
    if test.verbose:
        print("MF " + git_files)
    files = {}
    for filename in git_files.split():
        if filename == '':
            continue
        files[filename] = True
    return files


if not os.path.exists(test.root + "/.git"):
    test.skip("Not in a git repository")

### Must trim output before and after our file list
files = get_source_files()

for filename in sorted(files.keys()):
    filename = os.path.join(test.root, filename)
    if not os.path.exists(filename):  # git file might be deleted but not yet staged
        continue
    if not re.search(r'include.*verilated.*\.[ch].*$', filename):
        continue
    if re.search(r'gtkwave', filename):
        continue

    contents = test.file_contents(filename) + "\n\n"
    seen_setters = {}
    seen_getters = {}
    lineno = 0
    for line in contents.splitlines():
        lineno += 1
        if re.search(r'^\s*\/\/', line):  # skip commented lines
            continue
        #print("L "+line)
        if re.search(r'^class', line):
            seen_setters = {}
            seen_getters = {}
            if test.verbose:
                print("C " + line)
            continue
        m = re.search(r'\s*void\s+([a-zA-Z0-9_]+)\([a-zA-Z0-9_]+.*', line)
        if m:
            setter_name = m.group(1)
            if test.verbose:
                print("S " + setter_name + " " + line)
            seen_setters[setter_name] = True
        m = re.search(r'\s*[a-zA-Z0-9_]+\s+([a-zA-Z0-9_]+)\(\s*\)', line)
        if m:
            getter_name = m.group(1)
            if getter_name == "cycle" and re.search(r'verilated_sc_trace', filename):
                continue  # hardcoded check for cycle() which looks like a setter but isn't
            if test.verbose:
                print("G " + setter_name + " " + line)
            if getter_name in seen_setters and getter_name not in seen_getters:
                test.error(filename + ":" + str(lineno) + ": '" + getter_name +
                           "()' came after its setter; swap order")
                seen_getters[getter_name] = True

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('vlt')

stdout_filename = os.path.join(test.obj_dir, test.name + "__test.vpp")

test.compile(
    # Override default flags
    v_flags=[''],
    verilator_flags=["-E -P +incdir+t -Mdir", test.obj_dir],
    verilator_flags2=[''],
    verilator_flags3=[''],
    verilator_make_gmake=False,
    make_top_shell=False,
    make_main=False,
    stdout_filename=stdout_filename)

test.files_identical(stdout_filename, test.golden_filename)

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios("simulator")
test.top_filename = "t/t_sys_file_basic.v"

test.unlink_ok(test.obj_dir + "/t_sys_file_basic_test.log")

test.compile(
    verilator_flags2=["--coverage-expr"],
    # Build without cached objects, see bug363
    make_flags=["VM_PARALLEL_BUILDS=0"],
)

test.execute()
test.files_identical(test.obj_dir + "/t_sys_file_basic_test.log", test.golden_filename)

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('vlt')
test.pli_filename = "t/t_var_pinsizes.cpp"
test.top_filename = "t/t_var_pinsizes.v"

test.compile(
    verilator_flags2=["-sc --pins-sc-uint --pins-sc-biguint --trace-vcd --exe", test.pli_filename],
    make_main=False)


def hgrep(re):
    test.file_grep(os.path.join(test.obj_dir, test.vm_prefix + ".h"), re)


hgrep(r'sc_core::sc_in<bool>\s+&i1;')
hgrep(r'sc_core::sc_in<sc_dt::sc_uint<8>\s>\s+&i8;')
hgrep(r'sc_core::sc_in<sc_dt::sc_uint<16>\s>\s+&i16;')
hgrep(r'sc_core::sc_in<sc_dt::sc_uint<32>\s>\s+&i32;')
hgrep(r'sc_core::sc_in<sc_dt::sc_uint<64>\s>\s+&i64;')
hgrep(r'sc_core::sc_in<sc_dt::sc_biguint<65>\s>\s+&i65;')
hgrep(r'sc_core::sc_in<sc_dt::sc_biguint<128>\s>\s+&i128;')
hgrep(r'sc_core::sc_in<sc_dt::sc_bv<513>\s>\s+&i513;')
hgrep(r'sc_core::sc_in<sc_dt::sc_bv<1>\s>\s+&ibv1;')
hgrep(r'sc_core::sc_in<sc_dt::sc_bv<16>\s>\s+&ibv16;')

hgrep(r'sc_core::sc_out<bool>\s+&o1;')
hgrep(r'sc_core::sc_out<sc_dt::sc_uint<8>\s>\s+&o8;')
hgrep(r'sc_core::sc_out<sc_dt::sc_uint<16>\s>\s+&o16;')
hgrep(r'sc_core::sc_out<sc_dt::sc_uint<32>\s>\s+&o32;')
hgrep(r'sc_core::sc_out<sc_dt::sc_uint<64>\s>\s+&o64;')
hgrep(r'sc_core::sc_out<sc_dt::sc_biguint<65>\s>\s+&o65;')
hgrep(r'sc_core::sc_out<sc_dt::sc_biguint<128>\s>\s+&o128;')
hgrep(r'sc_core::sc_out<sc_dt::sc_bv<513>\s>\s+&o513;')
hgrep(r'sc_core::sc_out<sc_dt::sc_bv<1>\s>\s+&obv1;')
hgrep(r'sc_core::sc_out<sc_dt::sc_bv<16>\s>\s+&obv16;')

hgrep(r'sc_core::sc_in<sc_dt::sc_biguint<1>\s>\s+&ibu1;')
hgrep(r'sc_core::sc_in<sc_dt::sc_biguint<8>\s>\s+&ibu8;')
hgrep(r'sc_core::sc_in<sc_dt::sc_biguint<16>\s>\s+&ibu16;')
hgrep(r'sc_core::sc_in<sc_dt::sc_biguint<64>\s>\s+&ibu64;')
hgrep(r'sc_core::sc_in<sc_dt::sc_biguint<512>\s>\s+&ibu512;')
hgrep(r'sc_core::sc_out<sc_dt::sc_biguint<1>\s>\s+&obu1;')
hgrep(r'sc_core::sc_out<sc_dt::sc_biguint<8>\s>\s+&obu8;')
hgrep(r'sc_core::sc_out<sc_dt::sc_biguint<16>\s>\s+&obu16;')
hgrep(r'sc_core::sc_out<sc_dt::sc_biguint<64>\s>\s+&obu64;')
hgrep(r'sc_core::sc_out<sc_dt::sc_biguint<512>\s>\s+&obu512;')
hgrep(r'sc_core::sc_in<sc_dt::sc_biguint<601>\s>\s+&ibu601;')
hgrep(r'sc_core::sc_out<sc_dt::sc_biguint<601>\s>\s+&obu601;')

test.execute()

test.passes()



# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2026 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0


def run(test):
    variant, fmt = test.parse_name(r"t_trace_type_dupes_([a-z]+)_([a-z]+)")

    test.top_filename = "t/t_trace_type_dupes.v"
    test.golden_filename = test.py_filename.rpartition(fmt)[0] + fmt + ".out"

    flags = ["--cc", f"--trace-{fmt}"]
    match variant:
        case "default":
            pass
        case "structs":
            flags.extend(["--trace-structs", "--output-split-ctrace 10"])
        case _:
            test.error(f"Unhandled test variant '{variant}'")

    test.compile(verilator_flags2=flags)

    test.execute()

    test.trace_identical(test.trace_filename, test.golden_filename)

    test.passes()



# pylint: disable=C0103,C0114,C0116,C0301,E0402,W0622
#
# Configuration file for Verilator's Sphinx documentation builder.
# SPDX-FileCopyrightText: 2003-2026 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0
#
# This file only contains overridden options. For a full list:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
#
# ----------------------------------------------------------------------
# -- Path setup

import os
import re
import sys

sys.path.insert(0, os.path.abspath('./_ext'))

import sphinx_rtd_theme  # pylint: disable=wrong-import-position,


def get_vlt_version():
    filename = "../../Makefile"
    with open(filename, "r", encoding="utf8") as fh:
        for line in fh:
            match = re.search(r"PACKAGE_VERSION *= *([a-z0-9.]+) +([-0-9]+)", line)
            if match:
                return match.group(1), match.group(2)
            match = re.search(r"PACKAGE_VERSION *= *([a-z0-9.]+) +devel", line)
            if match:
                try:
                    data = os.popen('git log -n 1 --pretty=%cs').read()
                except Exception:  # pylint: disable=broad-except
                    data = ""  # fallback, and Sphinx will fill in today's date
                return "Devel " + match.group(1), data
    return "unknown", "unknown"


def setup(app):
    app.add_css_file('css/vlt_sphinx.css')


# ----------------------------------------------------------------------
# -- Project information

project = 'Verilator'
copyright = '2026 by Wilson Snyder, under LGPL-3.0 or Artistic-2.0'
author = 'Wilson Snyder'

# The master toctree document.
master_doc = "index"

version, today_fmt = get_vlt_version()
release = version

rst_prolog = """
.. role:: vlopt(option)
"""

# ----------------------------------------------------------------------
# -- General configuration

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
#
# To install:
#   sudo install enchant
#   python3 -m venv --system-site-packages ~/.verilator_pyenv
#   source ~/.verilator_pyenv/bin/activate
#   pip3 install sphinx sphinx_rtd_theme sphinxcontrib-spelling breathe
#
# We keep this extensions list empty for now to avoid needing dependencies
extensions = []
# extensions = ['breathe', 'sphinxcontrib.spelling']
# For website builds, this will add sphinxcontrib.jquery
if 'VERILATOR_SPHINX_EXTENSIONS' in os.environ:
    extensions.extend(os.environ.get('VERILATOR_SPHINX_EXTENSIONS', '').split(':'))

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    '_build', 'Thumbs.db', '.DS_Store', 'internals.rst', 'gen/ex_*', 'CONTRIBUTING.rst'
]

# Warn about refs
nitpicky = True
nitpicky_ignore = []

# Number figures for referencing
numfig = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
source_suffix = [".rst"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# Could use this for internals<->guide references
# intersphinx_mapping = { 'sphinx': ('https://sphinx-doc.org/', None), }

# ----------------------------------------------------------------------
# -- Options for HTML output

# html_baseurl =

html_domain_indices = False

html_logo = "../_static/verilator_192_150_min.png"

html_theme = 'sphinx_rtd_theme'
# Despite a 'WARNING: Calling get_html_theme_path is deprecated'
# the HTML output is different without this line
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

html_theme_options = {
    'logo_only': True,
    'style_nav_header_background': "#45acf8",  # Default is #2980B9
    # 'canonical_url':
}

html_context = {
    'display_github': True,
    'github_user': 'verilator',
    'github_repo': 'verilator',
    'github_version': 'master/docs/guide/',
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['../_static']

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = False

html_copy_source = False
html_show_sourcelink = False

html_use_index = False

html_favicon = "../_static/verilator_32x32_min.png"

# Custom sidebar templates, maps document names to template names.
# html_sidebars

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
# html_extra_path = []

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# ----------------------------------------------------------------------
# -- Options for Latex output

latex_logo = "../_static/verilator_logo.png"

latex_elements = {
    'extraclassoptions': 'openany,oneside',
    'papersize': 'letterpaper',
    'makeindex': '',
    'printindex': '',
    # 'pointsize': '10pt',
    # 'preamble': '',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
# latex_documents = [ ... ]

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

latex_domain_indices = False

# ----------------------------------------------------------------------
# -- Options for manual page output

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
# man_pages = [(master_doc, "verilog-to-routing", "Verilog-to-Routing Documentation", [author], 1)]

# If true, show URL addresses after external links.
# man_show_urls = False

# ----------------------------------------------------------------------
# -- Options for spelling

spelling_word_list_filename = ['spelling.txt']
spelling_ignore_contributor_names = True

# ----------------------------------------------------------------------
# -- Options for doxygen

# if shutil.which("doxygen"):
#    breathe_projects = {
#        "verilated": "../_build/doxygen/verilated/xml",
#    }
#    breathe_default_project = "verilated"
#
#    if not os.environ.get("SKIP_DOXYGEN", None) == "True":
#        for prjname, prjdir in breathe_projects.items():
#            print("Generating doxygen files for {}...".format(prjname))
#            os.makedirs(prjdir, exist_ok=True)
#            cmd = "cd ../_doxygen && doxygen {}.dox".format(prjname)
#            subprocess.call(cmd, shell=True)
#    else:
#        for prjname, prjdir in breathe_projects.items():
#            assert os.path.exists(prjdir) == True, "Regenerate doxygen XML for {}".format(prjname)

breathe_projects = {"verilated": "_build/doxygen/verilated/xml/"}
breathe_default_project = "verilated"
breathe_default_members = 'members'



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('simulator')
test.top_filename = "t/t_inst_tree.v"

default_vltmt_threads = test.get_default_vltmt_threads
test.compile(
    # Disable CFunc inlining so functions exist to be combined
    verilator_flags2=['--stats', '-fno-inline-cfuncs', test.t_dir + "/" + test.name + ".vlt"],
    # Force 3 threads even if we have fewer cores
    threads=(default_vltmt_threads if test.vltmt else 1))


def check_relative_refs(mod, expect_relative):
    found_relative = False
    for filename in test.glob_some(test.obj_dir + "/V" + test.name + "_" + mod + "*.cpp"):
        if test.verbose:
            print("FILE " + filename)
        text = test.file_contents(filename)

        if re.search(r'this->', text) or re.search(r'vlSelf->', text):
            if test.verbose:
                print(" REL " + filename)
            found_relative = True

        if found_relative != expect_relative:
            test.error(filename + " " +
                       ("has 'relative'" if found_relative else "has 'non-relative'") +
                       " variable references but expected " +
                       ("'relative'" if expect_relative else "'non-relative'"))


if test.vlt_all:
    # We expect to combine sequent functions across multiple instances of
    # l2, l3, l4, l5. If this number drops, please confirm this has not broken.
    test.file_grep(test.stats, r'Optimizations, Combined CFuncs\s+(\d+)',
                   (85 if test.vltmt else 67))

    # Everything should use relative references
    check_relative_refs("t", True)
    check_relative_refs("l1", True)
    check_relative_refs("l2", True)
    check_relative_refs("l3", True)
    check_relative_refs("l4", True)
    check_relative_refs("l5__P1", True)
    check_relative_refs("l5__P2", True)

test.execute()
test.file_grep(test.run_log_filename, r"\] (%m|.*t\.ps): Clocked")

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap
import filecmp

test.scenarios('simulator')
test.top_filename = "t/t_dpi_import.v"

tmp_dir = test.obj_dir + "/dpi-hdr"
test.mkdir_ok(tmp_dir)

test.compile(
    # Override default flags also
    verilator_flags=["-Wall -Wno-DECLFILENAME -Mdir " + tmp_dir + " --dpi-hdr-only"],
    verilator_make_gmake=False)

files = glob.glob(tmp_dir + "/*")

if len(files) < 1:
    test.error("Did not produce DPI header")
if len(files) > 1:
    test.error("Too many files created:" + ', '.join(files))

tmp_header = files[0]
print("============" + tmp_header)

if not re.search(r'__Dpi\.h$', tmp_header):
    test.error("Unexpected file " + tmp_header)

test.compile(verilator_flags2=["-Wall -Wno-DECLFILENAME"], verilator_make_gmake=False)

files = glob.glob(test.obj_dir + "/*__Dpi.h")
header = files[0]

if not filecmp.cmp(tmp_header, header):
    test.error("DPI header files are not the same")

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('simulator')

test.compile(make_top_shell=False,
             make_main=False,
             make_pli=True,
             iv_flags2=["-g2005-sv"],
             verilator_flags2=[
                 "--exe --vpi --public-flat-rw --no-l2name", test.pli_filename, "t/TestVpiMain.cpp"
             ],
             make_flags=['CPPFLAGS_ADD=-DVL_NO_LEGACY'])

test.execute(use_libvpi=True,
             expect_filename=test.golden_filename,
             xrun_run_expect_filename=re.sub(r'\.out$', '.xrun.out', test.golden_filename),
             iv_run_expect_filename=re.sub(r'\.out$', '.iv.out', test.golden_filename))

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('linter')

test.lint(verilator_flags2=["--bbox-unsup"], fails=test.vlt_all)

# Cannot use .out, get "$end" or "end of file" depending on bison version
test.file_grep(test.compile_log_filename, r"EOF in 'table'")

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('vlt_all')

# Use --debug-protect to assist debug

# This test makes randomly named .cpp/.h files, which tend to collect, so remove them first
for filename in (glob.glob(test.obj_dir + "/*_PS*.cpp") + glob.glob(test.obj_dir + "/*_PS*.h") +
                 glob.glob(test.obj_dir + "/*.d")):
    test.unlink_ok(filename)

test.compile(verilator_flags2=[
    "--protect-ids", "--protect-key SECRET_KEY", "--trace-vcd", "--coverage", "-Wno-INSECURE",
    "t/t_protect_ids_c.cpp"
])

test.execute()

# 'to="PS"' indicates means we probably mis-protected something already protected
# Use --debug-protect to assist debugging these
test.file_grep_not(test.obj_dir + "/" + test.vm_prefix + "__idmap.xml", r'to="PS')

if test.vlt_all:
    # Check for secret in any outputs
    for filename in test.glob_some(test.obj_dir + "/*.[ch]*"):
        if re.search(r'secret', filename, re.IGNORECASE):
            test.error("Secret found in a filename: " + filename)
        test.file_grep_not(filename, r'secret')

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('vlt')
test.top_filename = "t/t_flag_make_cmake.v"

test.compile(
    # Need --no-print-directory so golden file doesn't compare directory names
    verilator_flags2=[
        "--build --MAKEFLAGS --no-print-directory", " --MAKEFLAGS illegal-flag-to-fail-make"
    ],
    # Recursive make breaks the golden compare
    #expect_filename = test.golden_filename
    fails='any')  # make returns exit code 2

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('vlt_all', 'xsim')

if test.benchmark:
    test.sim_time = test.benchmark * 100

trace_opt = ""

secret_prefix = "secret"
secret_dir = test.obj_dir + "/" + secret_prefix
test.mkdir_ok(secret_dir)

# Always compile the secret file with Verilator no matter what simulator
#   we are testing with
test.run(
    logfile=secret_dir + "/vlt_compile.log",
    cmd=["perl", os.environ["VERILATOR_ROOT"] + "/bin/verilator",
         "-cc",
         trace_opt,
         "--prefix", "Vt_lib_prot_secret",
         "-Mdir", secret_dir,
         "--protect-lib", secret_prefix,
         "--protect-key", "secret-key",
         "t/t_lib_prot_comb.v"],
    verilator_run=True)  # yapf:disable

test.run(logfile=secret_dir + "/secret_gcc.log",
         cmd=[os.environ["MAKE"], "-C", secret_dir, "-f", "Vt_lib_prot_secret.mk"])

test.compile(
    verilator_flags2=[
        secret_dir + "/secret.sv",
        "+define+PROCESS_TOP",
        "-LDFLAGS",secret_prefix + "/libsecret.a"],
    xsim_flags2=[secret_dir + "/secret.sv"])  # yapf:disable

test.execute(xsim_run_flags2=["--sv_lib", secret_dir + "/libsecret", "--dpi_absolute"])

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('simulator')
test.top_filename = test.obj_dir + "/t_timing_long.v"

# Look for O(n^2) problems in process handling


def gen(filename):
    with open(filename, 'w', encoding="utf8") as fh:
        fh.write("// Generated by t_timing_long.py\n")

        fh.write("\n")
        fh.write("`ifdef TEST_VERBOSE\n")
        fh.write(" `define MSG(m) $display m\n")
        fh.write("`else\n")
        fh.write(" `define MSG(m)\n")
        fh.write("`endif\n")
        fh.write("\n")

        fh.write("module t;\n")
        fh.write("\n")
        fh.write("  int cnt;\n")
        fh.write("\n")
        fh.write("  initial begin\n")

        n = 100
        for i in range(1, n):
            # If statement around the timing is important to make the code scheduling
            # mostly unpredictable
            fh.write("    if (cnt == " + str(i - 1) + ") begin\n")
            fh.write("      #1; ++cnt; `MSG((\"[%0t] cnt?=" + str(i) + "\", $time));" +
                     " if (cnt != " + str(i) + ") $stop;\n")
            fh.write("    end\n")

        fh.write("\n")
        fh.write('    $write("*-* All Finished *-*\\n");' + "\n")
        fh.write("    $finish;\n")
        fh.write("  end\n")
        fh.write("endmodule\n")


gen(test.top_filename)

if test.have_coroutines:
    test.compile(verilator_flags2=["--binary"], make_top=1)

    test.execute()

test.compile(verilator_flags2=["--binary --no-timing -Wno-STMTDLY"], make_top=1)

test.execute()

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('simulator')
test.top_filename = "t/t_flag_make_cmake.v"

test.compile(verilator_flags2=['--make json'],
             verilator_make_gmake=False,
             verilator_make_cmake=False)

nout = test.run_capture("jq --version", check=False)
version_match = re.search(r'jq-([0-9.]+)', nout, re.IGNORECASE)
if not version_match:
    test.skip("jq is not installed")

json_filename = test.obj_dir + "/" + test.vm_prefix + ".json"
if not os.path.exists(json_filename):
    test.error(json_filename + " does not exist")

test.run(cmd=['cat "' + json_filename + '" | jq ".version"'])

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('vlt')
test.top_filename = "t/t_time_sc.v"

test.sc_time_resolution = 'SC_NS'

test.compile(verilator_flags2=[
    '-sc',
    '-timescale 1ps/1ps',  # Mismatch w/sc_time_resolution
    '+define+TEST_EXPECT=2us'
])

test.execute(fails=True, expect_filename=test.golden_filename)

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('vlt_all')


def check_splits():
    got1 = False
    gotSyms1 = False
    for filename in test.glob_some(test.obj_dir + "/*.cpp"):
        if re.search(r'Syms__.*__1', filename):
            gotSyms1 = True
        elif re.search(r'__1', filename):
            got1 = True
    if not got1:
        test.error("No __1 split file found")
    if not gotSyms1:
        test.error("No Syms__*__1 split file found")


test.compile(timing_loop=True,
             verilator_flags2=["--timing --output-split-cfuncs 1 -CFLAGS -Werror"])

test.execute()

if test.have_coroutines:
    check_splits()

test.passes()



# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2026 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import re


def covergroup_coverage_report(test, outfile=None):
    """Parse coverage.dat and write a sorted covergroup bin hit-count report.

    Lines have the form:  <hierarchy>[  [bin_type]]: <count>
    ignore_bins and illegal_bins are annotated with [ignore] / [illegal].

    Returns the path to the written report file.
    """
    if outfile is None:
        outfile = test.obj_dir + "/covergroup_report.txt"
    contents = test.file_contents(test.coverage_filename)
    entries = []
    for m in re.finditer(r"C '([^']+)' (\d+)", contents):
        entry, count = m.group(1), m.group(2)
        if '\x01t\x02covergroup' not in entry:
            continue
        h_m = re.search(r'\x01h\x02([^\x01]+)', entry)
        if not h_m:
            continue
        hier = h_m.group(1)
        bt_m = re.search(r'\x01bin_type\x02([^\x01]+)', entry)
        cross_m = re.search(r'\x01cross\x021', entry)
        annotations = []
        if bt_m:
            annotations.append(bt_m.group(1))
        if cross_m:
            annotations.append("cross")
        label = f"{hier} [{','.join(annotations)}]" if annotations else hier
        entries.append((hier, label, int(count)))
    entries.sort()
    with open(outfile, 'w', encoding='utf-8') as fh:
        for _hier, label, count in entries:
            fh.write(f"{label}: {count}\n")
    return outfile


def run(test, *, verilator_flags2=()):
    test.compile(verilator_flags2=['--coverage', *verilator_flags2])
    test.execute()
    covergroup_coverage_report(test)
    test.files_identical(test.obj_dir + '/covergroup_report.txt', test.golden_filename)
    test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

# show-config: This test runs the very first time we've executed Verilator
# after building so we make sure to run with --gdbbt, so if it dumps we'll
# get a trace.

import vltest_bootstrap

test.priority(100)
test.scenarios('vlt')

test.leak_check_disable()

DEBUG_QUIET = "--debug --debugi 0 --gdbbt --no-dump-tree"

test.run(
    cmd=[
        "perl",
        os.environ["VERILATOR_ROOT"] + "/bin/verilator",  #
        DEBUG_QUIET,
        "-V"
    ],
    verilator_run=True)

test.compile(verilator_flags2=[DEBUG_QUIET, "--trace-vcd"])

test.execute()

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('dist')

# Check all error messages match our standard format
# This assumes .out files cover all important errors


def formats():
    files = test.root + "/test_regress/t/*.out"
    warns = {}
    lnmatch = 0
    for filename in test.glob_some(files):
        if re.search(r'\.sarif\.out', filename):
            continue
        wholefile = test.file_contents(filename)
        filename = os.path.basename(filename)
        if re.search(r'(Exiting due to|%Error|%Warning)', wholefile):
            lineno = 0
            for line in wholefile.splitlines():
                lineno += 1
                line = re.sub(r'(\$display|\$write).*\".*%(Error|Warning)', '', line)
                line = re.sub(r'<---.*', '', line)
                if (re.search(r'(Error|Warning)', line)
                        and not re.search(r'^\s*<sformatf ', line)  # skip XML tag
                        and not re.search(r'^\s*{"type":"', line)  # skip JSON node
                        and not re.search(r'Error-internal-contents-bad', line)):
                    # These formats are documented in bin/verilator
                    # Error with fileline
                    # For testing only: we assume no : in filename
                    match = line
                    match = re.sub(r'^\[\d+\] ', '', match)  # Simplify runtime errors
                    if re.search(r'^%(Error|Warning)(-[A-Z0-9_]+)?: ([^:]+):(\d+):((\d+):)? ',
                                 match):
                        lnmatch += 1
                        if test.verbose:
                            print("ok-el " + filename + " " + line)
                    # Error no fileline
                    # For testing only: we assume any : is single quoted
                    elif re.search(r"^%(Error|Warning)(-[A-Z0-9_]+)?: [^:']+", match):
                        if test.verbose:
                            print("ok-en " + filename + " " + line)
                    else:
                        #print("FF "+filename+": "+line)
                        fline = filename + ":" + str(lineno)
                        warns[fline] = "Non-standard warning/error: " + fline + ": " + line
                        if test.verbose:
                            print(warns[fline])

    if not lnmatch:
        test.error("Check line number regexp is correct, no matches")
    if len(warns):
        # First warning lists everything as that's shown in the driver summary
        test.error_keep_going(' '.join(sorted(warns.keys())))
        for filename in sorted(warns.keys()):
            test.error_keep_going(warns[filename])


if not os.path.exists(test.root + "/.git"):
    test.skip("Not in a git repository")

formats()

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('simulator')


def gen(filename):
    # Generate using file to avoid missing newline in repository
    with open(filename, 'w', encoding="utf8") as fh:
        fh.write("// Generated by t_vthread.py\n")
        fh.write("1\n")
        fh.write("10\n")
        fh.write("20\n")
        fh.write("30")
        # No newline


gen(test.obj_dir + "/dat.mem")

test.compile()

test.execute()

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('vlt_all')
test.top_filename = "t/t_opt_merge_cond.v"

test.compile(verilator_flags2=["-unroll-count 64", "--stats", "-fno-merge-cond"])

test.execute()

if test.vlt:
    # Note, with vltmt this might be split differently, so only checking vlt
    test.file_grep_not(test.stats, r'Optimizations, MergeCond merges\s+(\d+)')
    test.file_grep_not(test.stats, r'Optimizations, MergeCond merged items\s+(\d+)')
    test.file_grep_not(test.stats, r'Optimizations, MergeCond longest merge\s+(\d+)')

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2026 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('vlt')

test.top_filename = "t/t_sched_ico_change_detect_input_assigned.v"

test.compile(make_top_shell=False,
             make_main=False,
             verilator_flags2=[
                 "--exe", "--stats", "-Wno-ASSIGNIN",
                 "t/t_sched_ico_change_detect_input_assigned.cpp", "-fno-ico-change-detect"
             ])

test.execute()

test.file_grep(test.stats, r"Scheduling, 'ico' change detect triggers\s+(\d+)", 0)

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('simulator')

test.compile(verilator_flags2=['-fno-gate'])

test.execute(expect_filename=test.golden_filename)

# Check for coverage on all POW functions
os.system("cat " + test.obj_dir + "/" + test.vm_prefix + "_*.cpp > " + test.obj_dir + "/all.cpp")

test.file_grep(test.obj_dir + "/all.cpp", r'VL_POW_III')
test.file_grep(test.obj_dir + "/all.cpp", r'VL_POW_IIQ')
test.file_grep(test.obj_dir + "/all.cpp", r'VL_POW_IIW')
test.file_grep(test.obj_dir + "/all.cpp", r'VL_POW_QQI')
test.file_grep(test.obj_dir + "/all.cpp", r'VL_POW_QQQ')
test.file_grep(test.obj_dir + "/all.cpp", r'VL_POW_QQW')
test.file_grep(test.obj_dir + "/all.cpp", r'VL_POW_WWI')
test.file_grep(test.obj_dir + "/all.cpp", r'VL_POW_WWQ')
test.file_grep(test.obj_dir + "/all.cpp", r'VL_POW_WWW')
test.file_grep(test.obj_dir + "/all.cpp", r'VL_POWSS_III')
test.file_grep(test.obj_dir + "/all.cpp", r'VL_POWSS_IIQ')
test.file_grep(test.obj_dir + "/all.cpp", r'VL_POWSS_IIW')
test.file_grep(test.obj_dir + "/all.cpp", r'VL_POWSS_QQI')
test.file_grep(test.obj_dir + "/all.cpp", r'VL_POWSS_QQQ')
test.file_grep(test.obj_dir + "/all.cpp", r'VL_POWSS_QQW')
test.file_grep(test.obj_dir + "/all.cpp", r'VL_POWSS_WWI')
test.file_grep(test.obj_dir + "/all.cpp", r'VL_POWSS_WWQ')
test.file_grep(test.obj_dir + "/all.cpp", r'VL_POWSS_WWW')

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This file ONLY is placed under the Creative Commons Public Domain, for
# any use, without warranty, 2025 by Wilson Snyder.
# SPDX-FileCopyrightText: 2025 Wilson Snyder
# SPDX-License-Identifier: CC0-1.0

import vltest_bootstrap

test.scenarios('vlt')

test.lint(expect_filename=test.golden_filename, fails=True)

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('dist')

test.run(cmd=[
    os.environ["VERILATOR_ROOT"] + "/bin/verilator_coverage",
    "--no-unlink",
    "--nounlink",
    "--write",
    test.obj_dir + "/coverage.dat",
    "t/t_vlcov_data_a.dat",
    "t/t_vlcov_data_b.dat",
    "t/t_vlcov_data_c.dat",
    "t/t_vlcov_data_d.dat",
],
         verilator_run=True)

# Not deleted e.g. parsed --no-unlink properly
test.files_identical(test.t_dir + "/t_vlcov_data_a.dat", test.t_dir + "/t_vlcov_data_a.dat")

# Older clib's didn't properly sort maps, but the coverage data doesn't
# really care about ordering. So avoid false failures by sorting.
test.files_identical_sorted(test.obj_dir + "/coverage.dat", test.golden_filename)

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2024 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('vlt')

if 'VERILATOR_TEST_NO_GDB' in os.environ:
    test.skip("Skipping due to VERILATOR_TEST_NO_GDB")
if not test.have_gdb:
    test.skip("No gdb installed")

test.lint(verilator_flags2=["--debug --gdbbt --debug-sigsegv"], fails='any')

test.file_grep(test.compile_log_filename, r'Program received signal SIGSEGV')
test.file_grep(test.compile_log_filename, r'in V3Options::')

test.passes()



#!/usr/bin/env python3
# DESCRIPTION: Verilator: Verilog Test driver/expect definition
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of either the GNU Lesser General Public License Version 3
# or the Perl Artistic License Version 2.0.
# SPDX-FileCopyrightText: 2025 Wilson Snyder
# SPDX-License-Identifier: LGPL-3.0-only OR Artistic-2.0

import vltest_bootstrap

test.scenarios('vlt')

# Compile with --binary --vpi to exercise the VPI-aware generated main.
# Also compile a VPI shared library to be loaded at runtime via +verilator+vpi+.
test.compile(make_pli=True, verilator_flags2=["--binary --vpi --public-flat-rw"])

# Run the generated binary; load the VPI library via the +verilator+vpi+ plusarg.
# The VPI library's output (observed 'count' reaching MAX_TICKS, then end-of-sim) is
# checked against the golden .out file.
# Also pass a non-VPI plusarg (skipped by the loader's prefix check) and a bare
# +verilator+vpi+ with an empty payload (skipped by the empty-arg check), so both
# loader-skip branches are exercised alongside the real library load.
test.execute(all_run_flags=[
    "+othertest", "+verilator+vpi+", "+verilator+vpi+" + test.obj_dir + "/libvpi.so"
],
             check_finished=True,
             expect_filename=test.golden_filename)

test.passes()
